﻿using System;
using Trinity.Config.Combat;
using Trinity.DbProvider;
using Trinity.Reference;
using Trinity.Technicals;
using Zeta.Common.Plugins;
using Zeta.Game;
using Zeta.Game.Internals.Actors;
using Zeta.Game.Internals.Actors.Gizmos;
using Logger = Trinity.Technicals.Logger;

namespace Trinity
{
    public partial class Trinity : IPlugin
    {
        private static bool RefreshGizmo(bool AddToCache)
        {
            if (!(c_diaObject is DiaGizmo))
                return false;

            c_diaGizmo = c_diaObject as DiaGizmo;

            if (!Settings.WorldObject.AllowPlayerResurection && CurrentCacheObject.ActorSNO == DataDictionary.SNOPlayerHeadStone)
            {
                c_InfosSubStep += "IgnoreHeadstones";
                AddToCache = false;
                return AddToCache;
            }
            // start as true, then set as false as we go. If nothing matches below, it will return true.
            AddToCache = true;

            bool openResplendentChest = CurrentCacheObject.InternalName.ToLower().Contains("chest_rare");

            // Ignore it if it's not in range yet, except shrines, pools of reflection and resplendent chests if we're opening chests
            if ((CurrentCacheObject.RadiusDistance > LootRange || CurrentCacheObject.RadiusDistance > 50) && CurrentCacheObject.Type != GObjectType.HealthWell &&
                CurrentCacheObject.Type != GObjectType.Shrine && CurrentCacheObject.RActorGuid != LastTarget.RActorGuid)
            {
                AddToCache = false;
                c_InfosSubStep += "NotInRange";
            }

            // re-add resplendent chests
            if (openResplendentChest)
            {
                AddToCache = true;
                c_InfosSubStep = "";
            }

            CacheObjectIsBountyObjective();

            CacheObjectMinimapActive();


            if (c_diaGizmo != null)
            {
                CurrentCacheObject.GizmoType = c_diaGizmo.ActorInfo.GizmoType;
            }

            // Anything that's been disabled by a script
            bool isGizmoDisabledByScript = false;
            try
            {
                if (c_diaObject is DiaGizmo)
                {
                    isGizmoDisabledByScript = c_diaGizmo.IsGizmoDisabledByScript;
                }
            }
            catch
            {
                // add methode & Dictionary check
                AddObjectToNavigationObstacleCache();

                Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement,
                    "Safely handled exception getting Gizmo-Disabled-By-Script attribute for object {0} [{1}]", CurrentCacheObject.InternalName, CurrentCacheObject.ActorSNO);
                c_InfosSubStep += "isGizmoDisabledByScriptException";
                //AddToCache = false;
            }
            if (isGizmoDisabledByScript)
            {
                // add methode & Dictionary check
                AddObjectToNavigationObstacleCache();

                AddToCache = false;
                c_InfosSubStep += "GizmoDisabledByScript";
                return AddToCache;

            }


            // Anything that's Untargetable
            bool untargetable = false;
            try
            {
                if (c_diaObject is DiaGizmo)
                {
                    untargetable = CacheObjectUntargetable() > 0;
                }
            }
            catch
            {

            }


            // Anything that's Invulnerable
            bool invulnerable = false;
            try
            {
                if (c_diaObject is DiaGizmo)
                {
                    invulnerable = CacheObjectInvulnerable() > 0;
                }
            }
            catch
            {

            }

            bool noDamage = false;
            try
            {
                if (c_diaObject is DiaGizmo)
                {
                    noDamage = CacheObjectNoDamage() > 0;
                }
            }
            catch
            {
                // add methode & Dictionary check
                AddObjectToNavigationObstacleCache();

                Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement,
                    "Safely handled exception getting NoDamage attribute for object {0} [{1}]", CurrentCacheObject.InternalName, CurrentCacheObject.ActorSNO);
                c_InfosSubStep += "NoDamage";
                //AddToCache = false;
            }

            double minDistance;
            bool gizmoUsed = false;
            switch (CurrentCacheObject.Type)
            {
                case GObjectType.Door:
                    {
                        AddToCache = true;

                        // add methode & Dictionary check
                        AddObjectToNavigationObstacleCache();

                        try
                        {
                            var gizmoDoor = c_diaObject as GizmoDoor;
                            if (gizmoDoor != null && gizmoDoor.IsLocked)
                            {
                                c_InfosSubStep += "IsLocked";
                                return false;
                            }
                        }
                        catch
                        {
                            c_InfosSubStep += "ErrorGettingIsLocked";
                        }

                        try
                        {
                            if (c_diaObject is DiaGizmo && ((DiaGizmo)c_diaObject).HasBeenOperated)
                            {
                                c_InfosSubStep += "Door has been operated";
                                return false;
                            }
                        }
                        catch
                        {
                            c_InfosSubStep += "ErrorGettingOperated";
                        }

                        try
                        {
                            string currentAnimation = c_diaObject.CommonData.CurrentAnimation.ToString().ToLower();
                            gizmoUsed = currentAnimation.EndsWith("open") || currentAnimation.EndsWith("opening");

                            // special hax for A3 Iron Gates
                            if (currentAnimation.Contains("irongate") && currentAnimation.Contains("open"))
                                gizmoUsed = false;
                            if (currentAnimation.Contains("irongate") && currentAnimation.Contains("idle"))
                                gizmoUsed = true;
                        }
                        catch 
                        {
                            c_InfosSubStep += "ErrorGettingOperated";
                        }

                        if (gizmoUsed)
                        {
                            c_InfosSubStep += "DoorUsed";
                            return false;
                        }

                        try
                        {
                            int gizmoState = CacheObjectGizmoState();
                            if (gizmoState == 1)
                            {
                                c_InfosSubStep += "GizmoState=1";
                                return false;
                            }
                        }
                        catch
                        {
                            c_InfosSubStep += "GizmoStateException";
                        }

                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            c_InfosSubStep += "Untargetable";
                            return false;
                        }


                        if (AddToCache)
                        {
                            try
                            {
                                DiaGizmo door = null;
                                if (c_diaObject is DiaGizmo)
                                {
                                    door = (DiaGizmo)c_diaObject;

                                    if (door != null && door.IsGizmoDisabledByScript)
                                    {
                                        // add methode & Dictionary check
                                        AddObjectToNavigationObstacleCache();

                                        c_InfosSubStep += "DoorDisabledbyScript";
                                        return false;
                                    }
                                }
                                else
                                {
                                    c_InfosSubStep += "InvalidCastToDoor";
                                    return false;
                                }
                            }

                            catch 
                            {
                                c_InfosSubStep += "InvalidDoor";
                                return false;
                            }
                        }
                    }
                    break;
                case GObjectType.Interactable:
                    {
                        AddToCache = true;

                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            c_InfosSubStep += "Untargetable";
                            return false;
                        }


                        int endAnimation;
                        if (DataDictionary.InteractEndAnimations.TryGetValue(CurrentCacheObject.ActorSNO, out endAnimation))
                        {
                            if (endAnimation == (int)c_diaGizmo.CommonData.CurrentAnimation)
                            {
                                c_InfosSubStep += "EndAnimation";
                                return false;
                            }
                        }

                        if (c_diaGizmo.GizmoState == 1)
                        {
                            c_InfosSubStep += "GizmoState1";
                            return false;
                        }

                        if (c_diaGizmo.HasBeenOperated)
                        {
                            c_InfosSubStep += "GizmoHasBeenOperated";
                            return false;
                        }

                        CurrentCacheObject.Radius = c_diaObject.CollisionSphere.Radius;
                    }
                    break;
                case GObjectType.HealthWell:
                    {
                        AddToCache = true;
                        try
                        {
                            gizmoUsed = (CacheObjectGizmoCharges() <= 0 && CacheObjectGizmoCharges() > 0);
                        }
                        catch
                        {
                            Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement, "Safely handled exception getting shrine-been-operated attribute for object {0} [{1}]", CurrentCacheObject.InternalName, CurrentCacheObject.ActorSNO);
                            AddToCache = true;
                            //return bWantThis;
                        }
                        try
                        {
                            int gizmoState = CacheObjectGizmoState();
                            if (gizmoState == 1)
                            {
                                AddToCache = false;
                                c_InfosSubStep += "GizmoState=1";
                                return AddToCache;
                            }
                        }
                        catch
                        {
                            AddToCache = false;
                            c_InfosSubStep += "GizmoStateException";
                            return AddToCache;
                        }
                        if (gizmoUsed)
                        {
                            c_InfosSubStep += "GizmoCharges";
                            AddToCache = false;
                            return AddToCache;
                        }
                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Untargetable";
                            return AddToCache;
                        }

                    }
                    break;
                case GObjectType.CursedShrine:
                case GObjectType.Shrine:
                    {
                        AddToCache = true;
                        // Shrines
                        // Check if either we want to ignore all shrines
                        if (!Settings.WorldObject.UseShrine)
                        {
                            // We're ignoring all shrines, so blacklist this one
                            c_InfosSubStep += "IgnoreAllShrinesSet";
                            AddToCache = false;
                            return AddToCache;
                        }

                        try
                        {
                            int gizmoState = CacheObjectGizmoState();
                            if (gizmoState == 1)
                            {
                                AddToCache = false;
                                c_InfosSubStep += "GizmoState=1";
                                return AddToCache;
                            }
                        }
                        catch
                        {
                            AddToCache = false;
                            c_InfosSubStep += "GizmoStateException";
                            return AddToCache;
                        }
                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Untargetable";
                            return AddToCache;
                        }


                        // Determine what shrine type it is, and blacklist if the user has disabled it
                        switch (CurrentCacheObject.ActorSNO)
                        {
                            case (int)SNOActor.Shrine_Global_Frenzied:  //Frenzy Shrine
                                if (!Settings.WorldObject.UseFrenzyShrine)
                                {
                                    TimedBlacklist.Add(CurrentTarget.RActorGuid, IgnoreDuringSeconds(60));
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    AddToCache = false;
                                }
                                if (Player.ActorClass == ActorClass.Monk && Settings.Combat.Monk.TROption.HasFlag(TempestRushOption.MovementOnly) && Skills.Monk.TempestRush.IsActive)
                                {
                                    // Frenzy shrines are a huge time sink for monks using Tempest Rush to move, we should ignore them.
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.Shrine_Global_Fortune:  //Fortune Shrine 
                                if (!Settings.WorldObject.UseFortuneShrine)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.Shrine_Global_Blessed: //Protection Shrine
                                if (!Settings.WorldObject.UseProtectionShrine)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.Shrine_Global_Reloaded: //Empowered Shrine - Shrine_Global_Reloaded
                                if (!Settings.WorldObject.UseEmpoweredShrine)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.Shrine_Global_Enlightened:  //Enlightened Shrine - Shrine_Global_Enlightened
                                if (!Settings.WorldObject.UseEnlightenedShrine)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.Shrine_Global_Hoarder:  //Fleeting Shrine
                                if (!Settings.WorldObject.UseFleetingShrine)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.x1_LR_Shrine_Infinite_Casting:  //Channeling Pylon - x1_LR_Shrine_Infinite_Casting
                                if (!Settings.WorldObject.UseChannelingPylon)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.x1_LR_Shrine_Electrified:  //Conduit Pylon - x1_LR_Shrine_Electrified
                                if (!Settings.WorldObject.UseConduitPylon)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.x1_LR_Shrine_Invulnerable:  //Shield Pylon -x1_LR_Shrine_Invulnerable
                                if (!Settings.WorldObject.UseShieldPylon)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;

                            case (int)SNOActor.x1_LR_Shrine_Run_Speed:  //Speed Pylon - x1_LR_Shrine_Run_Speed
                                if (!Settings.WorldObject.UseSpeedPylon)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;
                            case (int)SNOActor.x1_LR_Shrine_Damage:  //Power Pylon - x1_LR_Shrine_Damage
                                if (!Settings.WorldObject.UsePowerPylon)
                                {
                                    AddToCache = false;
                                    c_InfosSubStep += "IgnoreShrineOption";
                                    return AddToCache;
                                }
                                break;
                        }  //end switch

                        // Bag it!
                        CurrentCacheObject.Radius = 4f;
                        break;
                    }
                case GObjectType.Barricade:
                    {
                        AddToCache = true;

                        var gizmoDestructible = c_diaObject as GizmoDestructible;
                        if (gizmoDestructible != null && gizmoDestructible.HitpointsCurrentPct <= 0)
                        {
                            c_InfosSubStep += "HitPoints0";
                            return false;
                        }

                        if (noDamage)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "NoDamage";
                            return AddToCache;
                        }
                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Untargetable";
                            return AddToCache;
                        }


                        if (invulnerable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Invulnerable";
                            return AddToCache;
                        }

                        float maxRadiusDistance;

                        if (DataDictionary.DestructableObjectRadius.TryGetValue(CurrentCacheObject.ActorSNO, out maxRadiusDistance))
                        {
                            if (CurrentCacheObject.RadiusDistance < maxRadiusDistance)
                            {
                                AddToCache = true;
                                c_InfosSubStep = "";
                            }
                        }

                        if (DateTime.UtcNow.Subtract(PlayerMover.LastGeneratedStuckPosition).TotalSeconds <= 1)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                            break;
                        }

                        // Set min distance to user-defined setting
                        minDistance = Settings.WorldObject.DestructibleRange + CurrentCacheObject.Radius;

                        // This object isn't yet in our destructible desire range
                        if (minDistance <= 0 || CurrentCacheObject.RadiusDistance > minDistance)
                        {
                            c_InfosSubStep += "NotInBarricadeRange";
                            AddToCache = false;
                            return AddToCache;
                        }

                        break;
                    }
                case GObjectType.Destructible:
                    {
                        AddToCache = true;

                        var gizmoDestructible = c_diaObject as GizmoDestructible;
                        try
                        {
                        if (gizmoDestructible != null && gizmoDestructible.HitpointsCurrentPct <= 0)
                        {
                                c_InfosSubStep += "HitPoints0";
                            return false;
                        }
                        }
                        catch { }

                        if (noDamage)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "NoDamage";
                            return AddToCache;
                        }

                        if (invulnerable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Invulnerable";
                            return AddToCache;
                        }
                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Untargetable";
                            return AddToCache;
                        }


                        if (Player.ActorClass == ActorClass.Monk && Skills.Monk.TempestRush.IsActive && Skills.Monk.SweepingWind.TimeSinceUse <= 150)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "MonkTR";
                            break;
                        }

                        if (Player.ActorClass == ActorClass.Monk && Skills.Monk.SweepingWind.IsActive && Skills.Monk.SweepingWind.IsBuffActive)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "MonkSW";
                            break;
                        }

                        if (CurrentCacheObject.IsQuestMonster || CurrentCacheObject.IsMinimapActive)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                            break;
                        }

                        if (!DataDictionary.ForceDestructibles.Contains(CurrentCacheObject.ActorSNO) && Settings.WorldObject.DestructibleOption == DestructibleIgnoreOption.ForceIgnore)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "ForceIgnoreDestructibles";
                            break;
                        }

                        if (DateTime.UtcNow.Subtract(PlayerMover.LastGeneratedStuckPosition).TotalSeconds <= 1)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                            break;
                        }

                        // Set min distance to user-defined setting
                        minDistance = Settings.WorldObject.DestructibleRange;

                        // Only break destructables if we're stuck and using IgnoreNonBlocking
                        if (Settings.WorldObject.DestructibleOption == DestructibleIgnoreOption.DestroyAll)
                        {
                            minDistance += 12f;
                            AddToCache = true;
                            c_InfosSubStep = "";
                        }

                        float maxRadiusDistance;

                        if (DataDictionary.DestructableObjectRadius.TryGetValue(CurrentCacheObject.ActorSNO, out maxRadiusDistance))
                        {
                            if (CurrentCacheObject.RadiusDistance < maxRadiusDistance)
                            {
                                AddToCache = true;
                                c_InfosSubStep = "";
                            }
                        }
                        // Always add large destructibles within ultra close range
                        if (!AddToCache && CurrentCacheObject.Radius >= 10f && CurrentCacheObject.RadiusDistance <= 5.9f)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                            break;
                        }

                        // This object isn't yet in our destructible desire range
                        if (AddToCache && (minDistance <= 1 || CurrentCacheObject.RadiusDistance > minDistance) && PlayerMover.GetMovementSpeed() >= 1)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "NotInDestructableRange";
                        }
                        if (AddToCache && CurrentCacheObject.RadiusDistance <= 2f && DateTime.UtcNow.Subtract(PlayerMover.LastGeneratedStuckPosition).TotalMilliseconds > 500)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "NotStuck2";
                        }

                        // Add if we're standing still and bumping into it
                        if (CurrentCacheObject.RadiusDistance <= 2f && PlayerMover.GetMovementSpeed() <= 0)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                        }

                        if (CurrentCacheObject.RActorGuid == LastTarget.RActorGuid)
                        {
                            AddToCache = true;
                            c_InfosSubStep = "";
                        }

                        break;
                    }
                case GObjectType.CursedChest:
                case GObjectType.Container:
                    {
                        AddToCache = false;

                        bool isRareChest = CurrentCacheObject.InternalName.ToLower().Contains("chest_rare") || DataDictionary.ResplendentChestIds.Contains(CurrentCacheObject.ActorSNO);
                        bool isChest = (!isRareChest && CurrentCacheObject.InternalName.ToLower().Contains("chest")) ||
                            DataDictionary.ContainerWhiteListIds.Contains(CurrentCacheObject.ActorSNO); // We know it's a container but this is not a known rare chest
                        bool isCorpse = CurrentCacheObject.InternalName.ToLower().Contains("corpse");
                        bool isWeaponRack = CurrentCacheObject.InternalName.ToLower().Contains("rack");
                        bool isGroundClicky = CurrentCacheObject.InternalName.ToLower().Contains("ground_clicky");

                        // We want to do some vendoring, so don't open anything new yet
                        if (ForceVendorRunASAP)
                        {
                            AddToCache = false;
                            c_InfosSubStep += "ForceVendorRunASAP";
                        }

                        // Already open, blacklist it and don't look at it again
                        bool chestOpen;
                        try
                        {
                            chestOpen = CacheObjectIsChestOpen() > 0;
                        }
                        catch
                        {
                            Logger.Log(TrinityLogLevel.Debug, LogCategory.CacheManagement, "Safely handled exception getting container-been-opened attribute for object {0} [{1}]", CurrentCacheObject.InternalName, CurrentCacheObject.ActorSNO);
                            c_InfosSubStep += "ChestOpenException";
                            AddToCache = false;
                            return AddToCache;
                        }

                        // Check if chest is open
                        if (chestOpen)
                        {
                            // It's already open!
                            AddToCache = false;
                            c_InfosSubStep += "AlreadyOpen";
                            return AddToCache;
                        }

                        if (untargetable)
                        {
                            // add methode & Dictionary check
                            AddObjectToNavigationObstacleCache();

                            AddToCache = false;
                            c_InfosSubStep += "Untargetable";
                            return AddToCache;
                        }

                        // Resplendent chests have no range check
                        if (isRareChest && Settings.WorldObject.OpenRareChests)
                        {
                            AddToCache = true;
                            return AddToCache;
                        }

                        // Regular container, check range
                        if (CurrentCacheObject.RadiusDistance <= Settings.WorldObject.ContainerOpenRange)
                        {
                            if (isChest && Settings.WorldObject.OpenContainers)
                                return true;

                            if (isCorpse && Settings.WorldObject.InspectCorpses)
                                return true;

                            if (isGroundClicky && Settings.WorldObject.InspectGroundClicky)
                                return true;

                            if (isWeaponRack && Settings.WorldObject.InspectWeaponRacks)
                                return true;
                        }

                        if (CurrentCacheObject.IsQuestMonster)
                        {
                            AddToCache = true;
                            return AddToCache;
                        }

                        if (Settings.WorldObject.OpenAnyContainer)
                        {
                            AddToCache = true;
                            return AddToCache;
                        }

                        if (!isChest && !isCorpse && !isRareChest)
                        {
                            c_InfosSubStep += "InvalidContainer";
                        }
                        else
                        {
                            c_InfosSubStep += "IgnoreContainer";
                        }
                        break;
                    }
            }
            return AddToCache;
        }

        private static int CacheObjectIsChestOpen()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.ChestOpen);
        }

        private static int CacheObjectGizmoCharges()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.GizmoCharges);
        }

        private static int CacheObjectGizmoState()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.GizmoState);
        }

        private static int CacheObjectNoDamage()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.NoDamage);
        }

        private static int CacheObjectInvulnerable()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.Invulnerable);
        }

        private static int CacheObjectUntargetable()
        {
            return c_diaObject.CommonData.GetAttribute<int>(ActorAttributeType.Untargetable);
        }
    }
}
