using System;
using System.Collections.Generic;
using System.Linq;
using Trinity.Cache;
using Trinity.Combat.Abilities;
using Trinity.Config.Combat;
using Trinity.DbProvider;
using Trinity.Items;
using Trinity.LazyCache;
using Trinity.Reference;
using Trinity.Technicals;
using Zeta.Bot;
using Zeta.Bot.Profile.Common;
using Zeta.Bot.Settings;
using Zeta.Common;
using Zeta.Game;
using Zeta.Game.Internals.Actors;
using Zeta.Game.Internals.SNO;
using Logger = Trinity.Technicals.Logger;

namespace Trinity
{
    /// <summary>
    /// Prototype Weighting System
    /// </summary>
    public class Weighting
    {
        const double MaxWeight = 50000d;

        public static float CalculateWeight(TrinityObject trinityObject)
        {
            var weights = new List<Weight>();

            weights.Add(new Weight(5000, WeightMethod.Set, WeightReason.StartingWeight));

            weights.Add(new Weight(5000, WeightMethod.Add, WeightReason.AvoidanceNearby));

            weights.Add(new Weight(5000, WeightMethod.Add, WeightReason.IsCloseRange));

            weights.Add(new Weight(5000, WeightMethod.Add, WeightReason.HighPriorityShrine));

            if (IsHealthGlobeEmergency(trinityObject))
                weights.Add(new Weight(5000, WeightMethod.Add, WeightReason.HealthEmergency));

            weights.Add(new Weight(5000, WeightMethod.Add, WeightReason.HighPriorityContainer));

            weights.Add(new Weight(0, WeightMethod.Set, WeightReason.IgnoreElites, true));

            //trinityObject.WeightFactors = weights;

            var finalWeight = 0f;

            foreach (var w in weights)
            {
                switch (w.Method)
                {
                    case WeightMethod.Add:
                        finalWeight = finalWeight + w.Amount;
                        break;

                    case WeightMethod.Subtract:
                        finalWeight = finalWeight - w.Amount;
                        break;

                    case WeightMethod.Multiply:
                        finalWeight = finalWeight * w.Amount;
                        break;

                    case WeightMethod.Set:
                        finalWeight = w.Amount;
                        break;
                }

                if (w.IsFinal)
                    break;
            }

            return (float)Math.Max(finalWeight, MaxWeight);
        }

        private static bool IsHealthGlobeEmergency(TrinityObject trinityObject)
        {
            return (CacheManager.Me.CurrentHealthPct <= CombatBase.EmergencyHealthGlobeLimit ||
                   CacheManager.Me.PrimaryResourcePct <= CombatBase.HealthGlobeResource) &&
                   trinityObject.IsGlobe && Trinity.Settings.Combat.Misc.HiPriorityHG;
        }

        public struct Weight
        {
            public Weight(float amount, WeightMethod method, WeightReason reason, bool isFinal = false)
            {
                Amount = amount;
                Reason = reason;
                Method = method;
                IsFinal = isFinal;
            }

            public bool IsFinal;
            public WeightReason Reason;
            public WeightMethod Method;
            public float Amount;
        }

        public enum WeightMethod
        {
            None = 0,
            Multiply,
            Set,
            Add,
            Subtract
        }

        public enum WeightReason
        {
            None = 0,
            StartingWeight,
            AvoidanceNearby,
            IsCloseRange,
            IgnoreElites,
            HighPriorityContainer,
            HighPriorityShrine,
            HealthEmergency,
        }
    }
}