﻿/*using System;
using System.Linq;
using Trinity.Config.Combat;
using Trinity.DbProvider;
using Trinity.Reference;
using Trinity.Technicals;
using Zeta.Common;
using Zeta.Game;
using Zeta.Game.Internals.Actors;
using Logger = Trinity.Technicals.Logger;

namespace Trinity.Combat.Abilities
{
    public class DemonHunterCombat : CombatBase
    {
        #region Fields
        public static DemonHunterSetting DHSettings
        {
            get { return DemonHunterSettings; }
        }
        private static bool DemonHunter_HasNoPrimary()
        {
            return !(Hotbar.Contains(SNOPower.X1_DemonHunter_EvasiveFire) ||
                Hotbar.Contains(SNOPower.DemonHunter_Bolas) ||
                Hotbar.Contains(SNOPower.X1_DemonHunter_EntanglingShot) ||
                Hotbar.Contains(SNOPower.DemonHunter_Grenades) ||
                Hotbar.Contains(SNOPower.DemonHunter_HungeringArrow));
        }
        private static int MaxSentryCount
        {
            get
            {
                return 2 + (Legendary.BombardiersRucksack.IsEquipped ? 2 : 0) + (Passives.DemonHunter.CustomEngineering.IsActive ? 1 : 0);
            }
        }
        private static bool IsSentryOnTarget(int minCount = 1, float range = 55f)
        {
            if (!CacheData.SentryTurret.Any() || CurrentTarget == null)
                return false;

            return CacheData.SentryTurret.Count(s => s.Position.Distance2D(CurrentTarget.Position) <= range && CurrentTarget.IsInLineOfSightOfPoint(s.Position)) >= minCount;
        }
        private static bool IsSentryOnPosition(Vector3 loc, int minCount = 1, float range = 55f)
        {
            if (loc == new Vector3())
                return false;

            if (!CacheData.SentryTurret.Any())
                return false;

            return CacheData.SentryTurret.Count(s => s.Position.Distance2D(CurrentTarget.Position) <= range && NavHelper.CanRayCast(s.Position, loc, true)) >= minCount;
        }
        private static float rangedAttackRange = -1f;
        private static float RangedAttackRange
        {
            get
            {
                if (rangedAttackRange >= 0f)
                    return rangedAttackRange;

                if (Sets.EmbodimentOfTheMarauder.IsMaxBonusActive && IsSentryOnTarget(2))
                    rangedAttackRange = 0f;
                else
                    rangedAttackRange = DHSettings.RangedAttackRange;

                return rangedAttackRange;
            }
        }
        private static Vector3 LastZeiOfStoneLocation = new Vector3();
        #endregion

        /// <summary>
        /// Get combat/avoidance buff/power
        /// </summary>
        /// <returns></returns>
        public static TrinityPower GetPower()
        {
            TrinityPower power = null;

            // Destructible
            if (UseDestructiblePower)
            {
                return GetDemonHunterDestroyPower();
            }

            // Buffs
            if ((!Player.IsInCombat || UseOOCBuff) && !IsCurrentlyAvoiding)
            {
                power = GetBuffPower();
                if (!IsNull(power)) { return power; }
            }

            // In Combat, Avoiding
            if (IsCurrentlyAvoiding)
            {
                power = GetCombatAvoidancePower();
                if (!IsNull(power)) { return power; }
            }

            if (Player.IsRooted || Player.IsIncapacitated)
            {
                power = GetCombatAvoidancePower();
                if (!IsNull(power)) { return power; }
            }

            if (IsCurrentlyAvoiding)
                return DefaultPower;

            // Marauder cast/move routine
            if ((CurrentTarget != null || Player.IsInCombat) && Sets.EmbodimentOfTheMarauder.IsMaxBonusActive)
            {
                power = RunMarauderRoutine();
                if (!IsNull(power)) { return power; }
            }

            // In combat, Not Avoiding
            if (CurrentTarget != null)
            {
                power = GetCombatBuffPower();
                if (!IsNull(power)) { return power; }

                if (CurrentTarget.IsUnit)
                {
                    // Bastion Of Will require primary usage
                    if (Enemies.Nearby.Units.Any() && ShouldRefreshBastiansGeneratorBuff)
                    {
                        power = GetPrimaryPower();
                        if (!IsNull(power)) { return power; }
                    }

                    power = GetCombatPower();
                    if (!IsNull(power)) { return power; }

                    power = GetPrimaryPower();
                    if (!IsNull(power)) { return power; }
                }
            }

            // Default attacks
            return DefaultPower;
        }

        /// <summary>
        /// Return primary power
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetPrimaryPower()
        {
            if (Sets.EmbodimentOfTheMarauder.IsMaxBonusActive && AreaHasCastCriteria(RangedSkillsCastArea))
            {
                if (Skills.DemonHunter.EvasiveFire.IsActive && !Skills.DemonHunter.EvasiveFire.Cast(RangedSkillsCastArea.Position))
                    return new TrinityPower(SNOPower.X1_DemonHunter_EvasiveFire, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.HungeringArrow.IsActive && !Skills.DemonHunter.HungeringArrow.Cast(RangedSkillsCastArea.Position))
                    return new TrinityPower(SNOPower.DemonHunter_HungeringArrow, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.EntanglingShot.IsActive && !Skills.DemonHunter.EntanglingShot.Cast(RangedSkillsCastArea.Position))
                    return new TrinityPower(SNOPower.X1_DemonHunter_EntanglingShot, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.Bolas.IsActive && !Skills.DemonHunter.Bolas.Cast(RangedSkillsCastArea.Position))
                    return new TrinityPower(SNOPower.DemonHunter_Bolas, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.Grenade.IsActive && !Skills.DemonHunter.Grenade.Cast(RangedSkillsCastArea.Position))
                    return new TrinityPower(SNOPower.DemonHunter_Grenades, RangedAttackRange, RangedSkillsCastArea.Position);
            }
            else if (Sets.EmbodimentOfTheMarauder.IsMaxBonusActive && RangedSkillsCastArea != null && RangedSkillsCastArea.Position != Vector3.Zero)
            {
                if (Skills.DemonHunter.EvasiveFire.IsActive)
                    return new TrinityPower(SNOPower.X1_DemonHunter_EvasiveFire, RangedAttackRange, RangedSkillsCastArea.Position); ;

                if (Skills.DemonHunter.HungeringArrow.IsActive)
                    return new TrinityPower(SNOPower.DemonHunter_HungeringArrow, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.EntanglingShot.IsActive)
                    return new TrinityPower(SNOPower.X1_DemonHunter_EntanglingShot, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.Bolas.IsActive)
                    return new TrinityPower(SNOPower.DemonHunter_Bolas, RangedAttackRange, RangedSkillsCastArea.Position);

                if (Skills.DemonHunter.Grenade.IsActive)
                    return new TrinityPower(SNOPower.DemonHunter_Grenades, RangedAttackRange, RangedSkillsCastArea.Position);
            }

            var target = TargetUtil.GetClosestTarget(65f, useWeights: false);
            if (target != default(TrinityCacheObject))
            {
                // Evasive Fire
                if (Hotbar.Contains(SNOPower.X1_DemonHunter_EvasiveFire))
                {
                    float range = !ShouldRefreshBastiansGeneratorBuff && (Player.PrimaryResourceMissing > 5 || Sets.EmbodimentOfTheMarauder.IsMaxBonusActive)
                        ? RangedAttackRange
                        : 10f;

                    return new TrinityPower(SNOPower.X1_DemonHunter_EvasiveFire, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Hungering Arrow
                if (Hotbar.Contains(SNOPower.DemonHunter_HungeringArrow))
                {
                    return new TrinityPower(SNOPower.DemonHunter_HungeringArrow, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Entangling shot
                if (Skills.DemonHunter.EntanglingShot.IsActive)
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_EntanglingShot, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Bola Shot
                if (Hotbar.Contains(SNOPower.DemonHunter_Bolas))
                {
                    float range = RangedAttackRange > 50f ? 50f : RangedAttackRange;
                    return new TrinityPower(SNOPower.DemonHunter_Bolas, range, RangedSkillsCastArea.Position);
                }

                // Grenades
                if (Hotbar.Contains(SNOPower.DemonHunter_Grenades))
                {
                    float range = RangedAttackRange > 30f ? 30f : RangedAttackRange;
                    return new TrinityPower(SNOPower.DemonHunter_Grenades, range, RangedSkillsCastArea.Position);
                }
            }

            var generator = SkillUtils.Active.FirstOrDefault(s => s.IsGeneratorOrPrimary);
            return generator != null ? generator.ToPower(RangedAttackRange, Enemies.BestCluster.Position) : DefaultPower;
        }

        /// <summary>
        /// Gets the best (non-movement related) avoidance power
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetCombatAvoidancePower()
        {
            // Smoke Screen
            if (CanCast(SNOPower.DemonHunter_SmokeScreen, CanCastFlags.NoTimer) &&
                !GetHasBuff(SNOPower.DemonHunter_ShadowPower))
            {
                return new TrinityPower(SNOPower.DemonHunter_SmokeScreen);
            }

            // Smoke Screen
            if (CanCast(SNOPower.DemonHunter_SmokeScreen, CanCastFlags.NoTimer) &&
                !GetHasBuff(SNOPower.DemonHunter_ShadowPower) &&
                (Player.CurrentHealthPct <= 0.50 || Player.IsRooted || TargetUtil.AnyMobsInRange(12) ||
                (Legendary.MeticulousBolts.IsEquipped && TargetUtil.AnyMobsInRange(60)) || Player.IsIncapacitated))
            {
                return new TrinityPower(SNOPower.DemonHunter_SmokeScreen);
            }

            return null;
        }

        /// <summary>
        /// Gets the best combat buff power for the current conditions
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetCombatBuffPower()
        {
            // NotSpam Shadow Power
            if (!DemonHunterSettings.SpamShadowPower && CanCast(SNOPower.DemonHunter_ShadowPower) && !Player.IsIncapacitated &&
                (!GetHasBuff(SNOPower.DemonHunter_ShadowPower) || Player.CurrentHealthPct <= EmergencyHealthPotionLimit) && // if we don't have the buff or our health is low
                (Player.CurrentHealthPct < 1f || Player.IsRooted || TargetUtil.AnyMobsInRange(15)))
            {
                return new TrinityPower(SNOPower.DemonHunter_ShadowPower);
            }

            // Smoke Screen
            if (CanCast(SNOPower.DemonHunter_SmokeScreen, CanCastFlags.NoTimer) &&
                !GetHasBuff(SNOPower.DemonHunter_ShadowPower) &&
                (Player.CurrentHealthPct <= 0.50 || Player.IsRooted || TargetUtil.AnyMobsInRange(12) ||
                (Legendary.MeticulousBolts.IsEquipped && TargetUtil.AnyMobsInRange(60)) || Player.IsIncapacitated))
            {
                return new TrinityPower(SNOPower.DemonHunter_SmokeScreen);
            }

            // Preparation, restore Disc if needed
            float useDelay = Runes.DemonHunter.FocusedMind.IsActive ? 15000 : 500;
            if (CanCast(SNOPower.DemonHunter_Preparation, CanCastFlags.NoTimer) &&
                Player.SecondaryResource <= V.F("DemonHunter.MinPreparationDiscipline") &&
                !Runes.DemonHunter.Punishment.IsActive &&
                TimeSincePowerUse(SNOPower.DemonHunter_Preparation) >= useDelay)
            {
                return new TrinityPower(SNOPower.DemonHunter_Preparation);
            }

            // Preparation: Punishment
            if (CanCast(SNOPower.DemonHunter_Preparation, CanCastFlags.NoTimer) &&
                Runes.DemonHunter.Punishment.IsActive && (Player.PrimaryResourceMax - Player.PrimaryResource) >= 75)
            {
                return new TrinityPower(SNOPower.DemonHunter_Preparation);
            }

            // Companion
            if (!Player.IsIncapacitated && Hotbar.Contains(SNOPower.X1_DemonHunter_Companion))
            {
                // Use Spider Slow on 4 or more trash mobs in an area or on Unique/Elite/Champion
                if (Runes.DemonHunter.SpiderCompanion.IsActive && CanCast(SNOPower.X1_DemonHunter_Companion) && TargetUtil.ClusterExists(25f, 4) && TargetUtil.EliteOrTrashInRange(25f))
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
                }

                //Use Bat when Hatred is Needed
                if (Runes.DemonHunter.BatCompanion.IsActive && CanCast(SNOPower.X1_DemonHunter_Companion) && Player.PrimaryResourceMissing >= 60)
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
                }

                // Use Boar Taunt on 3 or more trash mobs in an area or on Unique/Elite/Champion
                if (Runes.DemonHunter.BoarCompanion.IsActive && CanCast(SNOPower.X1_DemonHunter_Companion) && ((TargetUtil.ClusterExists(20f, 4) && TargetUtil.EliteOrTrashInRange(20f)) ||
                    (CurrentTarget.IsBossOrEliteRareUnique && CurrentTarget.Distance <= 20f)))
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
                }

                // Ferrets used for picking up Health Globes when low on Health
                if (Runes.DemonHunter.FerretCompanion.IsActive && Trinity.ObjectCache.Any(o => o.Type == GObjectType.HealthGlobe && o.Distance < 60f) && Player.CurrentHealthPct < EmergencyHealthPotionLimit)
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
                }

                // Use Wolf Howl on Unique/Elite/Champion - Would help for farming trash, but trash farming should not need this - Used on Elites to reduce Deaths per hour
                if (Runes.DemonHunter.WolfCompanion.IsActive && CanCast(SNOPower.X1_DemonHunter_Companion, CanCastFlags.NoTimer) &&
                    ((CurrentTarget.IsBossOrEliteRareUnique || TargetUtil.AnyMobsInRange(40, 10)) && CurrentTarget.RadiusDistance < 25f))
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
                }
            }

            // Companion active attack on elite
            if (CanCast(SNOPower.X1_DemonHunter_Companion) && CurrentTarget.IsEliteRareUnique)
            {
                return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
            }

            // Companion off CD
            if (CanCast(SNOPower.X1_DemonHunter_Companion, CanCastFlags.NoTimer) && TargetUtil.AnyMobsInRange(60) && DemonHunterSettings.CompanionOffCooldown)
            {
                return new TrinityPower(SNOPower.X1_DemonHunter_Companion);
            }

            int mfdResource = Hotbar.Contains(SNOPower.DemonHunter_SmokeScreen) ? (Runes.DemonHunter.MortalEnemy.IsActive ? 3 : 17) : 17;
            // Marked for Death
            if (CanCast(SNOPower.DemonHunter_MarkedForDeath) &&
                Player.SecondaryResource >= mfdResource &&
                !CurrentTarget.HasDebuff(SNOPower.DemonHunter_MarkedForDeath))
            {
                return new TrinityPower(SNOPower.DemonHunter_MarkedForDeath, DHSettings.RangedAttackRange, CurrentTarget.Position);
            }

            // Vengeance
            if (CanCast(SNOPower.X1_DemonHunter_Vengeance, CanCastFlags.NoTimer) &&
                ((!DemonHunterSettings.VengeanceElitesOnly && TargetUtil.AnyMobsInRange(DHSettings.RangedAttackRange + 5f, 5)) ||
                TargetUtil.AnyElitesInRange(DHSettings.RangedAttackRange + 5f)))
            {
                if (Runes.DemonHunter.Seethe.IsActive && Player.PrimaryResourcePct < 0.3 &&
                    (!Runes.DemonHunter.Punishment.IsActive || !CanCast(SNOPower.DemonHunter_Preparation, CanCastFlags.NoTimer)) &&
                    (!Runes.DemonHunter.BatCompanion.IsActive || !Sets.EmbodimentOfTheMarauder.IsFirstBonusActive || !CanCast(SNOPower.X1_DemonHunter_Companion, CanCastFlags.NoTimer)))
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Vengeance);
                }
                else if (!Runes.DemonHunter.Seethe.IsActive)
                {
                    return new TrinityPower(SNOPower.X1_DemonHunter_Vengeance);
                }
            }

            // Rain of Vengeance OffCD with Dark Cloud
            if (!Player.IsInTown && CanCast(SNOPower.DemonHunter_RainOfVengeance, CanCastFlags.NoTimer) &&
               !Player.IsIncapacitated && Runes.DemonHunter.DarkCloud.IsActive && DemonHunterSettings.RainOfVengeanceOffCD)
            {
                return new TrinityPower(SNOPower.DemonHunter_RainOfVengeance);
            }

            return null;
        }

        /// <summary>
        /// Gets the best combat power for the current conditions
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetCombatPower()
        {
            // Rain of Vengeance
            if (Skills.DemonHunter.RainOfVengeance.CanCast() && !Player.IsIncapacitated &&
                (TargetUtil.ClusterExists(45f, 5) || TargetUtil.EliteOrTrashInRange(45f) || DemonHunterSettings.RainOfVengeanceOffCD || Sets.NatalyasVengeance.IsEquipped))
            {
                return new TrinityPower(SNOPower.DemonHunter_RainOfVengeance, 0f, Enemies.BestLargeCluster.Position);
            }

            // Sentry Turret
            if (SentryCastArea != null && CanCast(SNOPower.DemonHunter_Sentry, CanCastFlags.NoTimer) && Trinity.PlayerOwnedDHSentryCount < MaxSentryCount)
            {
                Vector3 zeiOfStoneNewTarget = MathEx.CalculatePointFrom(Player.Position, SentryCastArea.Position, 51f);
                if (IsZeisOfStoneEquipped && !Runes.DemonHunter.PolarStation.IsActive && zeiOfStoneNewTarget.Distance2D(LastZeiOfStoneLocation) >= 25f &&
                    SentryCastArea.Units.OrderByDescending(u => u.Distance).FirstOrDefault().IsInLineOfSightOfPoint(zeiOfStoneNewTarget))
                {
                    LastZeiOfStoneLocation = zeiOfStoneNewTarget;
                    return new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, zeiOfStoneNewTarget);
                }
                else
                    return new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, SentryCastArea.Position);
            }

            // Caltrops
            if (CanCast(SNOPower.DemonHunter_Caltrops) && TargetUtil.AnyMobsInRange(40) && !GetHasBuff(SNOPower.DemonHunter_Caltrops))
            {
                return new TrinityPower(SNOPower.DemonHunter_Caltrops);
            }

            // Fan of Knives
            if (CanCast(SNOPower.DemonHunter_FanOfKnives) &&
                (TargetUtil.EliteOrTrashInRange(15) || TargetUtil.AnyTrashInRange(15f, 5, false)))
            {
                return new TrinityPower(SNOPower.DemonHunter_FanOfKnives);
            }

            // Strafe spam - similar to barbarian whirlwind routine
            if (CanCast(SNOPower.DemonHunter_Strafe) &&
                Player.PrimaryResource >= DemonHunterSettings.StrafeMinHatred)
            {
                bool shouldGetNewZigZag =
                    (DateTime.UtcNow.Subtract(LastChangedZigZag).TotalMilliseconds >= V.I("Barbarian.Whirlwind.ZigZagMaxTime") ||
                    CurrentTarget.ACDGuid != LastZigZagUnitAcdGuid ||
                    ZigZagPosition.Distance2D(Player.Position) <= 5f);

                if (shouldGetNewZigZag)
                {
                    var wwdist = V.F("Barbarian.Whirlwind.ZigZagDistance");

                    ZigZagPosition = TargetUtil.GetZigZagTarget(CurrentTarget.Position, wwdist);

                    LastZigZagUnitAcdGuid = CurrentTarget.ACDGuid;
                    LastChangedZigZag = DateTime.UtcNow;
                }

                int postCastTickDelay = TrinityPower.MillisecondsToTickDelay(250);

                return new TrinityPower(SNOPower.DemonHunter_Strafe, 15f, ZigZagPosition, Trinity.Player.WorldDynamicID, -1, 0, postCastTickDelay);
            }

            // Spike Trap
            if (CanCast(SNOPower.DemonHunter_SpikeTrap) &&
                LastPowerUsed != SNOPower.DemonHunter_SpikeTrap)
            {
                // For distant monsters, try to target a little bit in-front of them (as they run towards us), if it's not a treasure goblin
                float reducedDistance = 0f;
                if (CurrentTarget.Distance > 17f && !CurrentTarget.IsTreasureGoblin)
                {
                    reducedDistance = CurrentTarget.Distance - 17f;
                    if (reducedDistance > 5f)
                        reducedDistance = 5f;
                }
                Vector3 vNewTarget = MathEx.CalculatePointFrom(CurrentTarget.Position, Player.Position, CurrentTarget.Distance - reducedDistance);
                return new TrinityPower(SNOPower.DemonHunter_SpikeTrap, 35f, vNewTarget, Trinity.Player.WorldDynamicID, -1, 1, 1);
            }

            if (AreaHasCastCriteria(RangedSkillsCastArea))
            {
                if (CurrentTarget.MonsterAffixes.HasFlag(MonsterAffixes.ReflectsDamage) &&
                    Sets.EmbodimentOfTheMarauder.IsFullyEquipped && IsSentryOnTarget())
                {
                    RangedSkillsCastArea.Position = MathEx.CalculatePointFrom(RangedSkillsCastArea.Position, Player.Position, -30f);
                }

                // Cluster Arrow
                if (CanCast(SNOPower.DemonHunter_ClusterArrow))
                {
                    if (!Sets.NatalyasVengeance.IsFullyEquipped || CacheData.Buffs.HasBuff(SNOPower.P2_ItemPassive_Unique_Ring_053))
                    {
                        return new TrinityPower(SNOPower.DemonHunter_ClusterArrow, RangedAttackRange, RangedSkillsCastArea.Position);
                    }
                }

                // Multi Shot
                if (CanCast(SNOPower.DemonHunter_Multishot))
                {
                    return new TrinityPower(SNOPower.DemonHunter_Multishot, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Elemental Arrow for non-kridershot
                if (CanCast(SNOPower.DemonHunter_ElementalArrow))
                {
                    return new TrinityPower(SNOPower.DemonHunter_ElementalArrow, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Chakram normal attack
                if (CanCast(SNOPower.DemonHunter_Chakram) && !Runes.DemonHunter.ShurikenCloud.IsActive)
                {
                    if (DHSettings.RangedAttackRange > 0)
                        DHSettings.RangedAttackRange = 50;

                    return new TrinityPower(SNOPower.DemonHunter_Chakram, RangedAttackRange, RangedSkillsCastArea.Position);
                }

                // Chakram:Shuriken Cloud
                if (!Player.IsInTown && CanCast(SNOPower.DemonHunter_Chakram, CanCastFlags.NoTimer) &&
                    Runes.DemonHunter.ShurikenCloud.IsActive && TimeSincePowerUse(SNOPower.DemonHunter_Chakram) >= 110000)
                {
                    return new TrinityPower(SNOPower.DemonHunter_Chakram);
                }

                // Impale
                if (CanCast(SNOPower.DemonHunter_Impale))
                {
                    return new TrinityPower(SNOPower.DemonHunter_Impale, RangedAttackRange, RangedSkillsCastArea.Position);
                }
            }

            // Rapid Fire
            if (CanCast(SNOPower.DemonHunter_RapidFire))
            {
                // Players with grenades *AND* rapid fire should spam grenades at close-range instead
                if (CanCast(SNOPower.DemonHunter_Grenades) && CurrentTarget.RadiusDistance <= 18f)
                {
                    return new TrinityPower(SNOPower.DemonHunter_Grenades, 18f, CurrentTarget.Position);
                }
                // Now return rapid fire, if not sending grenades instead
                return new TrinityPower(SNOPower.DemonHunter_RapidFire, DHSettings.RangedAttackRange, CurrentTarget.Position);
            }

            var spender = SkillUtils.Active.FirstOrDefault(s => s.IsAttackSpender && s.CanCast());
            return spender != null ? spender.ToPower(RangedAttackRange, Enemies.BestCluster.Position) : null;
        }

        /// <summary>
        /// Checks and casts buffs if needed
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetBuffPower()
        {
            // Shadow Power on low health (nospam)
            if (CanCast(SNOPower.DemonHunter_ShadowPower) && !Player.IsIncapacitated &&
                (TimeSincePowerUse(SNOPower.DemonHunter_ShadowPower) > 4500 || !GetHasBuff(SNOPower.DemonHunter_ShadowPower)) &&
                Player.CurrentHealthPct <= Trinity.PlayerEmergencyHealthPotionLimit &&
                Player.SecondaryResource >= 14) // When spamming Shadow Power, save some Discipline for emergencies
            {
                return new TrinityPower(SNOPower.DemonHunter_ShadowPower);
            }

            // Spam Shadow Power
            if (DemonHunterSettings.SpamShadowPower && CanCast(SNOPower.DemonHunter_ShadowPower) && !Player.IsIncapacitated &&
                TimeSincePowerUse(SNOPower.DemonHunter_ShadowPower) > 250 && !GetHasBuff(SNOPower.DemonHunter_ShadowPower) &&
                Player.SecondaryResource >= 14)
            {
                return new TrinityPower(SNOPower.DemonHunter_ShadowPower);
            }

            // Smoke Screen spam
            if (DemonHunterSettings.SpamSmokeScreen && CanCast(SNOPower.DemonHunter_SmokeScreen) &&
                !GetHasBuff(SNOPower.DemonHunter_ShadowPower))
            {
                return new TrinityPower(SNOPower.DemonHunter_SmokeScreen);
            }

            // Preparation, restore Disc if needed
            float useDelay = Runes.DemonHunter.FocusedMind.IsActive ? 15000 : 500;
            if (CanCast(SNOPower.DemonHunter_Preparation, CanCastFlags.NoTimer) &&
            Player.SecondaryResource <= V.F("DemonHunter.MinPreparationDiscipline") &&
            !Runes.DemonHunter.Punishment.IsActive &&
            TimeSincePowerUse(SNOPower.DemonHunter_Preparation) >= useDelay)
            {
                return new TrinityPower(SNOPower.DemonHunter_Preparation);
            }

            // Preparation: Punishment
            if (CanCast(SNOPower.DemonHunter_Preparation, CanCastFlags.NoTimer) &&
                Runes.DemonHunter.Punishment.IsActive && (Player.PrimaryResourceMax - Player.PrimaryResource) >= 75)
            {
                return new TrinityPower(SNOPower.DemonHunter_Preparation);
            }

            return null;
        }

        /// <summary>
        /// Get best power to destroy object
        /// </summary>
        /// <returns></returns>
        private static TrinityPower GetDemonHunterDestroyPower()
        {
            if (Hotbar.Contains(SNOPower.DemonHunter_HungeringArrow))
                return new TrinityPower(SNOPower.DemonHunter_HungeringArrow, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.X1_DemonHunter_EntanglingShot))
                return new TrinityPower(SNOPower.X1_DemonHunter_EntanglingShot, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.DemonHunter_Bolas))
                return new TrinityPower(SNOPower.DemonHunter_Bolas, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.DemonHunter_Grenades))
                return new TrinityPower(SNOPower.DemonHunter_Grenades, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.DemonHunter_ElementalArrow) && Player.PrimaryResource >= 10)
                return new TrinityPower(SNOPower.DemonHunter_ElementalArrow, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.X1_DemonHunter_EvasiveFire))
                return new TrinityPower(SNOPower.X1_DemonHunter_EvasiveFire, 10f, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.DemonHunter_RapidFire) && Player.PrimaryResource >= 10)
                return new TrinityPower(SNOPower.DemonHunter_RapidFire, DHSettings.RangedAttackRange, CurrentTarget.Position);

            if (Hotbar.Contains(SNOPower.DemonHunter_Chakram) && Player.PrimaryResource >= 20)
                return new TrinityPower(SNOPower.DemonHunter_Chakram, 50f, CurrentTarget.Position);

            return DefaultPower;
        }

        /// <summary>
        /// Special marauder combat routine
        /// </summary>
        /// <returns></returns>
        public static TrinityPower RunMarauderRoutine()
        {
            if (CurrentTarget == null)
                return null;

            if (CurrentTarget.IsUnit)
                SwitchToTarget(TargetUtil.GetClosestTarget(150f));

            // Fields
            TrinityPower power = null;

            power = GetCombatBuffPower();
            if (!IsNull(power)) { return power; }

            // Sentry
            if (SentryCastArea != null && Skills.DemonHunter.Sentry.CanCast(CanCastFlags.NoTimer) && Trinity.PlayerOwnedDHSentryCount < MaxSentryCount)
            {
                Vector3 zeiOfStoneNewTarget = MathEx.CalculatePointFrom(Player.Position, SentryCastArea.Position, 51f);
                if (AreaHasCastCriteria(SentryCastArea, true))
                {
                    if (IsZeisOfStoneEquipped && !Runes.DemonHunter.PolarStation.IsActive && zeiOfStoneNewTarget.Distance2D(LastZeiOfStoneLocation) >= 25f &&
                        SentryCastArea.Units.OrderByDescending(u => u.Distance).FirstOrDefault().IsInLineOfSightOfPoint(zeiOfStoneNewTarget))
                    {
                        if (!Skills.DemonHunter.Sentry.Cast(zeiOfStoneNewTarget))
                        {
                            LastZeiOfStoneLocation = zeiOfStoneNewTarget;
                            power = new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, zeiOfStoneNewTarget);
                        }
                    }
                    else if (!Skills.DemonHunter.Sentry.Cast(SentryCastArea.Position))
                    {
                        power = new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, SentryCastArea.Position);
                    }
                }
                else if (SentryCastArea != null && SentryCastArea.Position != Vector3.Zero)
                {
                    if (IsZeisOfStoneEquipped && !Runes.DemonHunter.PolarStation.IsActive && zeiOfStoneNewTarget.Distance2D(LastZeiOfStoneLocation) >= 25f &&
                        SentryCastArea.Units.OrderByDescending(u => u.Distance).FirstOrDefault().IsInLineOfSightOfPoint(zeiOfStoneNewTarget))
                    {
                        LastZeiOfStoneLocation = zeiOfStoneNewTarget;
                        power = new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, zeiOfStoneNewTarget);
                    }
                    else
                    {
                        power = new TrinityPower(SNOPower.DemonHunter_Sentry, DHSettings.RangedAttackRange, SentryCastArea.Position);
                    }
                }
            }

            // SentryCastSkills
            if (AreaHasCastCriteria(RangedSkillsCastArea))
            {
                if (Skills.DemonHunter.ClusterArrow.CanCast() && !Skills.DemonHunter.ClusterArrow.Cast(RangedSkillsCastArea.Position))
                    power = new TrinityPower(SNOPower.DemonHunter_ClusterArrow, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (CanCast(SNOPower.DemonHunter_Multishot, CanCastFlags.NoTimer) && !Skills.DemonHunter.Multishot.Cast(RangedSkillsCastArea.Position))
                    power = new TrinityPower(SNOPower.DemonHunter_Multishot, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.Chakram.CanCast() && !Skills.DemonHunter.Chakram.Cast(RangedSkillsCastArea.Position))
                    power = new TrinityPower(SNOPower.DemonHunter_Chakram, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.Impale.CanCast() && !Skills.DemonHunter.Impale.Cast(RangedSkillsCastArea.Position))
                    power = new TrinityPower(SNOPower.DemonHunter_Impale, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.ElementalArrow.CanCast() && !Skills.DemonHunter.ElementalArrow.Cast(RangedSkillsCastArea.Position))
                    power = new TrinityPower(SNOPower.DemonHunter_ElementalArrow, RangedAttackRange, RangedSkillsCastArea.Position);
            }
            else if (IsNull(power) && RangedSkillsCastArea != null && RangedSkillsCastArea.Position != Vector3.Zero)
            {
                if (Skills.DemonHunter.ClusterArrow.CanCast(CanCastFlags.NoTimer))
                    power = new TrinityPower(SNOPower.DemonHunter_ClusterArrow, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (CanCast(SNOPower.DemonHunter_Multishot, CanCastFlags.NoTimer))
                    power = new TrinityPower(SNOPower.DemonHunter_Multishot, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.Chakram.CanCast(CanCastFlags.NoTimer))
                    power = new TrinityPower(SNOPower.DemonHunter_Chakram, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.Impale.CanCast(CanCastFlags.NoTimer))
                    power = new TrinityPower(SNOPower.DemonHunter_Impale, RangedAttackRange, RangedSkillsCastArea.Position);
                else if (Skills.DemonHunter.ElementalArrow.CanCast(CanCastFlags.NoTimer))
                    power = new TrinityPower(SNOPower.DemonHunter_ElementalArrow, RangedAttackRange, RangedSkillsCastArea.Position);
            }

            return power;
        }

        #region Marauder routine helpers
        public static void ResetArea()
        {
            sentryCastArea = null;
            rangedSkillsCastArea = null;
            rangedAttackRange = -1f;
        }

        private static TargetArea sentryCastArea { get; set; }
        private static TargetArea SentryCastArea
        {
            get
            {
                if (sentryCastArea != null)
                    return sentryCastArea;

                TrinityCacheObject targetCacheObject = TargetUtil.GetClosestTarget(150f, Player.Position, false);

                if (Runes.DemonHunter.PolarStation.IsActive && TargetUtil.ClusterExists(20f, DHSettings.RangedAttackRange, 3))
                    sentryCastArea = Enemies.BestCluster;

                else if (TargetUtil.ClusterExists(20f, DHSettings.RangedAttackRange))
                    sentryCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, DHSettings.RangedAttackRange + 5f, _useWeights: false).Position);

                else if (TargetUtil.ClusterExists(20f, 65f))
                    sentryCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, 65f + 5f, _useWeights: false).Position);

                else if (CurrentTarget != null && CurrentTarget.IsUnit)
                    sentryCastArea = new TargetArea(60f, CurrentTarget.Position);

                else if (targetCacheObject != null && targetCacheObject != default(TrinityCacheObject) && targetCacheObject.Position != Vector3.Zero)
                    sentryCastArea = new TargetArea(60f, targetCacheObject.Position);

                return sentryCastArea;
            }
        }

        private static TargetArea rangedSkillsCastArea { get; set; }
        private static TargetArea RangedSkillsCastArea
        {
            get
            {
                if (rangedSkillsCastArea != null)
                    return rangedSkillsCastArea;

                TrinityCacheObject targetCacheObject = default(TrinityCacheObject);
                if (Skills.DemonHunter.ElementalArrow.IsActive && Runes.DemonHunter.BallLightning.IsActive)
                    targetCacheObject = TargetUtil.GetBestPierceTarget(DHSettings.RangedAttackRange);
                else
                    targetCacheObject = TargetUtil.GetClosestTarget(150f, useWeights: false);

                if (Skills.DemonHunter.ClusterArrow.IsActive && TargetUtil.ClusterExists(20f, DHSettings.RangedAttackRange))
                    rangedSkillsCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, DHSettings.RangedAttackRange + 5f, _useWeights: false).Position);

                else if (Skills.DemonHunter.Multishot.IsActive && TargetUtil.ClusterExists(40f, DHSettings.RangedAttackRange))
                    rangedSkillsCastArea = new TargetArea(40f, TargetUtil.GetBestClusterUnit(40f, DHSettings.RangedAttackRange + 5f, _useWeights: false).Position);

                else if (Skills.DemonHunter.Chakram.IsActive && TargetUtil.ClusterExists(20f, Math.Min(DHSettings.RangedAttackRange, 50f)))
                    rangedSkillsCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, Math.Min(DHSettings.RangedAttackRange + 5f, 50f), _useWeights: false).Position);

                if (Skills.DemonHunter.ClusterArrow.IsActive && TargetUtil.ClusterExists(20f, 90f))
                    rangedSkillsCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, 65f, _useWeights: false).Position);

                else if (Skills.DemonHunter.Multishot.IsActive && TargetUtil.ClusterExists(40f, 90f))
                    rangedSkillsCastArea = new TargetArea(40f, TargetUtil.GetBestClusterUnit(40f, 65f, _useWeights: false).Position);

                else if (Skills.DemonHunter.Chakram.IsActive && TargetUtil.ClusterExists(20f, 90f))
                    rangedSkillsCastArea = new TargetArea(20f, TargetUtil.GetBestClusterUnit(20f, 65f, _useWeights: false).Position);

                else if (targetCacheObject != null && targetCacheObject != default(TrinityCacheObject) && targetCacheObject.Position != Vector3.Zero)
                    rangedSkillsCastArea = new TargetArea(20f, targetCacheObject.Position);

                else if (CurrentTarget != null && CurrentTarget.IsUnit)
                    rangedSkillsCastArea = new TargetArea(20f, CurrentTarget.Position);
                return rangedSkillsCastArea;
            }
        }

        private static bool AreaHasCastCriteria(TargetArea area, bool rangeRequired = false)
        {
            return (area != null && area.Position != Vector3.Zero && area.UnitCount >= 1 &&
                (area.Position.Distance2D(Trinity.Player.Position) <= DHSettings.RangedAttackRange &&
                area.Units.OrderByDescending(u => u.Distance).FirstOrDefault().IsInLineOfSight ||
                RangedAttackRange <= 1f && !rangeRequired));
        }

        #endregion
        #region Vault helpers
        public static bool CurrentlyUseVault
        {
            get
            {
                return Player.ActorClass == ActorClass.DemonHunter && Hotbar.Contains(SNOPower.DemonHunter_Vault) &&
                    CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) < 400;
            }
        }

        public static bool CanCastCombatVaultMovement
        {
            get
            {
                return CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= DemonHunterSettings.VaultCombatDelay &&
                    Skills.DemonHunter.Vault.CanCast() &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.MovementOnly;
            }
        }

        public static bool CanCastOocVaultMovement
        {
            get
            {
                return CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= DemonHunterSettings.VaultMovementDelay &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.CombatOnly &&
                    !TargetUtil.AnyMobsInRange(85f, false) &&
                    Skills.DemonHunter.Vault.CanCast() &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.CombatOnly &&
                    ((Player.SecondaryResourcePct > DemonHunterSettings.MinDiciplineOOCVaultPct && !Sets.DanettasHatred.IsFullyEquipped) ||
                    Sets.DanettasHatred.IsFullyEquipped) && !Passives.DemonHunter.TacticalAdvantage.IsBuffActive;
            }
        }

        public static bool IsVaultAptCombatMovement(Vector3 loc)
        {
            float minCombatDist = DemonHunterSettings.MinDistVaultShrineOther;
            if (CurrentTarget != null) { minCombatDist += CurrentTarget.RequiredRange; }
            float minDicipline = 0f;
            bool inLoS = true;
            bool specialTimeSinceUse = true;
            if (CurrentTarget != null)
            {
                inLoS = CurrentTarget.IsInLineOfSight;
                switch (CurrentTarget.Type)
                {
                    case GObjectType.Player:
                        return false;
                    case GObjectType.Unit:
                        minCombatDist = DemonHunterSettings.MinDistVaultCombat +
                            Math.Max(CurrentTarget.RequiredRange, CombatBase.KiteDistance) +
                            TargetUtil.GetClosestTarget(150f).Position.Distance2D(CurrentTarget.Position);
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 2000;
                        break;
                    case GObjectType.ProgressionGlobe:
                    case GObjectType.PowerGlobe:
                    case GObjectType.Item:
                    case GObjectType.MarkerLocation:
                    case GObjectType.Gold:
                        minCombatDist = DemonHunterSettings.MinDistVaultGlobeItem + CurrentTarget.RequiredRange;
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1500;
                        break;
                    case GObjectType.HealthWell:
                    case GObjectType.HealthGlobe:
                        if (Player.AvoidDeath || Player.CurrentHealthPct < 0.4)
                        {
                            minCombatDist = 0f;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 100;
                        }
                        else if (Player.CurrentHealthPct < 0.6)
                        {
                            minCombatDist = DemonHunterSettings.MinDistVaultHealth + CurrentTarget.RequiredRange;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1000;
                        }
                        else
                        {
                            minCombatDist = DemonHunterSettings.MinDistVaultGlobeItem + CurrentTarget.RequiredRange;
                            minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1500;
                        }
                        break;
                    default:
                        minCombatDist = DemonHunterSettings.MinDistVaultShrineOther + CurrentTarget.RequiredRange;
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1000;
                        break;
                }
            }
            else
            {
                minCombatDist = DemonHunterSettings.MinDistVaultCombat + CombatBase.KiteDistance;
                minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 2000;
            }

            return
                // Real combat movement 
                !Trinity.Player.StandingInAvoidance && !Trinity.Player.AvoidDeath && !Trinity.Player.NeedToKite &&
                // Move to target & target in LoS 
                PlayerMover.IsMovementToTarget(loc) && inLoS &&
                // Save dicipline for avoid vault movement 
                Player.SecondaryResourcePct > minDicipline &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= minCombatDist && loc.Distance2D(Trinity.Player.Position) < 150 && specialTimeSinceUse &&
                // target not in 0.8 kite monster zone 
                !Trinity.ObjectCache.Any(o => o.IsUnit && loc.Distance2D(o.Position) <= CombatBase.KiteDistance * 0.8) &&
                // Loc not in aoe and not intersect aoe 
                !TargetUtil.LocOrPathInAoE(loc);
        }

        public static bool IsVaultAptKiteMovement(Vector3 loc)
        {
            return
                // Vault only at min dh settings health AND 
                Player.CurrentHealthPct <= DHSettings.MinHealthVaultKite &&
                // ha can fight at loc 
                (CurrentTarget == null || (CurrentTarget.IsInLineOfSightOfPoint(loc) && loc.Distance2D(Trinity.Player.Position) < MainGrid.MinRangeToTarget)) &&

                // 1) A real kite movement OR 
                ((!Trinity.Player.StandingInAvoidance && !Trinity.Player.AvoidDeath && Trinity.Player.NeedToKite) ||

                // 2) Target not null and not move to target 
                (CurrentTarget != null && CurrentTarget.IsUnit && !PlayerMover.IsMovementToTarget(loc) && CurrentTarget.Distance <= CombatBase.KiteDistance &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultKite &&
                // Loc not in aoe and not intersect aoe OR 
                !TargetUtil.LocOrPathInAoE(loc)) ||

                // 3) Target is kite 
                (CurrentTarget != null && CurrentTarget.IsKite &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultKite));
        }

        public static bool IsVaultAptAvoidanceMovement(Vector3 loc)
        {
            return Player.CurrentHealthPct <= DHSettings.MinHealthVaultAvoidance &&
                (Trinity.Player.AvoidDeath || CombatBase.PlayerIsSurrounded || Trinity.Player.StandingInAvoidance ||
                (CurrentTarget != null && CurrentTarget.IsAvoidance && !CurrentTarget.IsKite)) &&
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultAvoidance;
        }

        public static void LogVault(Vector3 loc, bool cMove = false, bool kMove = false, bool aMove = false, string wI = "")
        {
            string log = "";
            if (aMove)
                log = string.Format("Using vault for avoidance movement{1}, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position),
                    CacheData.AvoidanceObstacles.Any() ?
                    " (" + CacheData.AvoidanceObstacles.OrderBy(a => Vector3.Distance(a.Position, Trinity.Player.Position)).FirstOrDefault().Name + ") " : "");

            else if (kMove)
                log = string.Format("Using vault for kite movement, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            else if (cMove)
                log = string.Format("Using vault for combat movement{1}, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position),
                    CurrentTarget != null ?
                    " (" + CurrentTarget.Type.ToString() + " Attempt) " : "");

            else if (!TargetUtil.AnyMobsInRange(85f, false))
                log = string.Format("Using vault for ooc movement, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            else
                log = string.Format("Using vault for unknown reason ?_?, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, log + (Settings.Advanced.LogCategories.HasFlag(LogCategory.Movement) && wI != "" ? " WeightsInfos:" + wI : ""));
        }
        #endregion
    }
}*/
using System;
using System.Collections.Generic;
using System.Linq;
using Trinity.Config.Combat;
using Trinity.Helpers;
using Trinity.Objects;
using Trinity.Reference;
using Zeta.Common;
using Zeta.Game.Internals.Actors;
using Trinity.Technicals;
using Logger = Trinity.Technicals.Logger;
using Zeta.Game;
using Trinity.DbProvider;

namespace Trinity.Combat.Abilities
{
    public class DemonHunterCombat : CombatBase
    {
        #region Fields
		/// <summary>
        /// Maximum number of sentries allowed from Equipped items and Passives
        /// </summary>
        public static int MaxSentryCount
        {
            get { return 2 + (Legendary.BombardiersRucksack.IsEquipped ? 2 : 0) + (Passives.DemonHunter.CustomEngineering.IsActive ? 1 : 0); }
        }

        /// <summary>
        /// If they are a sentry on target
        /// </summary>
        private static bool IsSentryOnTarget(int minCount = 1, float range = 55f)
        {
            if (!CacheData.SentryTurret.Any() || CurrentTarget == null)
                return false;
            return CacheData.SentryTurret.Count(s => s.Position.Distance2D(CurrentTarget.Position) <= range && CurrentTarget.IsInLineOfSightOfPoint(s.Position)) >= minCount;
        }

        /// <summary>
        /// If they are a sentry on target
        /// </summary>
        private static bool IsSentryOnPosition(Vector3 loc, int minCount = 1, float range = 55f)
        {
            if (loc == new Vector3() || !CacheData.SentryTurret.Any())
                return false;
            return CacheData.SentryTurret.Count(s => s.Position.Distance2D(CurrentTarget.Position) <= range && NavHelper.CanRayCast(s.Position, loc, true)) >= minCount;
        }

        /// <summary>
        /// Return the ranged attack range set by the user
        /// </summary>
        private static float RangedAttackRange
        {
            get
            {
                if (Sets.EmbodimentOfTheMarauder.IsMaxBonusActive && IsSentryOnTarget(2))
                    return 0f;
                return DemonHunterSettings.RangedAttackRange;
            }
        }

        /// <summary>
        /// Return the last sentry cast position with Zei's Of Stone logic
        /// </summary>
        private static Vector3 LastZeiOfStoneLocation = new Vector3(); 
	    #endregion

        #region Vault helpers
        public static bool CurrentlyUseVault
        {
            get
            {
                return Player.ActorClass == ActorClass.DemonHunter && Skills.DemonHunter.Vault.IsActive &&
                    TimeSincePowerUse(SNOPower.DemonHunter_Vault) < 400;
            }
        }

        public static bool CanCastCombatVaultMovement
        {
            get
            {
                return CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= DemonHunterSettings.VaultCombatDelay &&
                    Skills.DemonHunter.Vault.CanCast() &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.MovementOnly;
            }
        }

        public static bool CanCastOocVaultMovement
        {
            get
            {
                return CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= DemonHunterSettings.VaultMovementDelay &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.CombatOnly &&
                    !TargetUtil.AnyMobsInRange(85f, false) &&
                    Skills.DemonHunter.Vault.CanCast() &&
                    DemonHunterSettings.VaultMode != DemonHunterVaultMode.CombatOnly &&
                    ((Player.SecondaryResourcePct > DemonHunterSettings.MinDiciplineOOCVaultPct && !Sets.DanettasHatred.IsFullyEquipped) ||
                    Sets.DanettasHatred.IsFullyEquipped) && !Passives.DemonHunter.TacticalAdvantage.IsBuffActive;
            }
        }

        public static bool IsVaultAptCombatMovement(Vector3 loc)
        {
            float minCombatDist = DemonHunterSettings.MinDistVaultShrineOther;
            if (CurrentTarget != null) { minCombatDist += CurrentTarget.RequiredRange; }
            float minDicipline = 0f;
            bool inLoS = true;
            bool specialTimeSinceUse = true;
            if (CurrentTarget != null)
            {
                inLoS = CurrentTarget.IsInLineOfSight;
                switch (CurrentTarget.Type)
                {
                    case GObjectType.Player:
                        return false;
                    case GObjectType.Unit:
                        minCombatDist = DemonHunterSettings.MinDistVaultCombat +
                            Math.Max(CurrentTarget.RequiredRange, CombatBase.KiteDistance) +
                            TargetUtil.GetClosestTarget(150f).Position.Distance2D(CurrentTarget.Position);
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 2000;
                        break;
                    case GObjectType.ProgressionGlobe:
                    case GObjectType.PowerGlobe:
                    case GObjectType.Item:
                    case GObjectType.MarkerLocation:
                    case GObjectType.Gold:
                        minCombatDist = DemonHunterSettings.MinDistVaultGlobeItem + CurrentTarget.RequiredRange;
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1500;
                        break;
                    case GObjectType.HealthWell:
                    case GObjectType.HealthGlobe:
                        if (Player.AvoidDeath || Player.CurrentHealthPct < 0.4)
                        {
                            minCombatDist = 0f;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 100;
                        }
                        else if (Player.CurrentHealthPct < 0.6)
                        {
                            minCombatDist = DemonHunterSettings.MinDistVaultHealth + CurrentTarget.RequiredRange;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1000;
                        }
                        else
                        {
                            minCombatDist = DemonHunterSettings.MinDistVaultGlobeItem + CurrentTarget.RequiredRange;
                            minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                            specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1500;
                        }
                        break;
                    default:
                        minCombatDist = DemonHunterSettings.MinDistVaultShrineOther + CurrentTarget.RequiredRange;
                        minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                        specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 1000;
                        break;
                }
            }
            else
            {
                minCombatDist = DemonHunterSettings.MinDistVaultCombat + CombatBase.KiteDistance;
                minDicipline = DemonHunterSettings.MinDiciplineCombatVaultPct;
                specialTimeSinceUse = CombatBase.TimeSincePowerUse(SNOPower.DemonHunter_Vault) >= 2000;
            }

            return
                // Real combat movement 
                !Trinity.Player.StandingInAvoidance && !Trinity.Player.AvoidDeath && !Trinity.Player.NeedToKite &&
                // Move to target & target in LoS 
                PlayerMover.IsMovementToTarget(loc) && inLoS &&
                // Save dicipline for avoid vault movement 
                Player.SecondaryResourcePct > minDicipline &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= minCombatDist && loc.Distance2D(Trinity.Player.Position) < 150 && specialTimeSinceUse &&
                // target not in 0.8 kite monster zone 
                !Trinity.ObjectCache.Any(o => o.IsUnit && loc.Distance2D(o.Position) <= CombatBase.KiteDistance * 0.8) &&
                // Loc not in aoe and not intersect aoe 
                !TargetUtil.LocOrPathInAoE(loc);
        }

        public static bool IsVaultAptKiteMovement(Vector3 loc)
        {
            return
                // Vault only at min dh settings health AND 
                Player.CurrentHealthPct <= DemonHunterSettings.MinHealthVaultKite && 
                // ha can fight at loc 
                (CurrentTarget == null || (CurrentTarget.IsInLineOfSightOfPoint(loc) && loc.Distance2D(Trinity.Player.Position) < MainGrid.MinRangeToTarget)) &&

                // 1) A real kite movement OR 
                ((!Trinity.Player.StandingInAvoidance && !Trinity.Player.AvoidDeath && Trinity.Player.NeedToKite) ||

                // 2) Target not null and not move to target 
                (CurrentTarget != null && CurrentTarget.IsUnit && !PlayerMover.IsMovementToTarget(loc) && CurrentTarget.Distance <= CombatBase.KiteDistance &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultKite &&
                // Loc not in aoe and not intersect aoe OR 
                !TargetUtil.LocOrPathInAoE(loc)) ||

                // 3) Target is kite 
                (CurrentTarget != null && CurrentTarget.IsKite &&
                // Distance to target respect dh settings 
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultKite));
        }

        public static bool IsVaultAptAvoidanceMovement(Vector3 loc)
        {
            return Player.CurrentHealthPct <= DemonHunterSettings.MinHealthVaultAvoidance &&
                (Trinity.Player.AvoidDeath || CombatBase.PlayerIsSurrounded || Trinity.Player.StandingInAvoidance ||
                (CurrentTarget != null && CurrentTarget.IsAvoidance && !CurrentTarget.IsKite)) &&
                loc.Distance2D(Trinity.Player.Position) >= DemonHunterSettings.MinDistVaultAvoidance;
        }

        public static void LogVault(Vector3 loc, bool cMove = false, bool kMove = false, bool aMove = false, string wI = "")
        {
            string log = "";
            if (aMove)
                log = string.Format("Using vault for avoidance movement{1}, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position),
                    CacheData.AvoidanceObstacles.Any() ?
                    " (" + CacheData.AvoidanceObstacles.OrderBy(a => Vector3.Distance(a.Position, Trinity.Player.Position)).FirstOrDefault().Name + ") " : "");

            else if (kMove)
                log = string.Format("Using vault for kite movement, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            else if (cMove)
                log = string.Format("Using vault for combat movement{1}, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position),
                    CurrentTarget != null ?
                    " (" + CurrentTarget.Type.ToString() + " Attempt) " : "");

            else if (!TargetUtil.AnyMobsInRange(85f, false))
                log = string.Format("Using vault for ooc movement, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            else
                log = string.Format("Using vault for unknown reason ?_?, Dist={0:0.0}", Vector3.Distance(loc, Trinity.Player.Position));

            Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, log + (Settings.Advanced.LogCategories.HasFlag(LogCategory.Movement) && wI != "" ? " WeightsInfos:" + wI : ""));
        }
        #endregion

        static DemonHunterCombat()
        {
            SkillUtils.SetSkillMeta(SkillsDefaultMeta.DemonHunter.ToList());
            SetConditions();
        }

        internal static void SetCombatSettings()
        {
            EmergencyHealthPotionLimit = DemonHunterSettings.PotionLevel;
            EmergencyHealthGlobeLimit = DemonHunterSettings.HealthGlobeLevel;
            HealthGlobeResource = DemonHunterSettings.HealthGlobeLevelResource;
            KiteDistance = DemonHunterSettings.KiteLimit;
            KiteMode = DemonHunterSettings.KiteMode;
        }
        
        /// <summary>
        /// Main method for selecting a power
        /// </summary>
        public static TrinityPower GetPower()
        {
            TrinityPower power;

            if (UseDestroyablePower && TryGetPower(GetDestructablesSkill(), out power))
                return power;

            if (UseAvoidancePower && TryGetPower(GetAvoidanceSkill(), out power))
                return power;

            if (UseOOCPower)
            {
                // Out of Combat Buffs
                if (!UseAvoidancePower && !Player.IsInTown && TryGetPower(GetBuffSkill(), out power))
                    return power;
            }
            else if (UseCombatPower)
            {
                // Use Generator for the Bastians Ring Set buff
                if (ShouldRefreshBastiansGeneratorBuff)
                {
                    SwitchToTarget(TargetUtil.GetClosestTarget());
                    if (TryGetPower(GetAttackGenerator(), out power))
                        return power;
                }

                // Use Spender for the Bastians Ring Set buff
                if (ShouldRefreshBastiansSpenderBuff)
                {
                    SwitchToTarget(TargetUtil.GetClosestTarget());
                    if (TryGetPower(GetAttackSpender(), out power))
                        return power;
                }

                if (!Skills.DemonHunter.Strafe.IsActive)
                    SwitchToTarget(TargetUtil.GetClosestTarget());

                // Main ability selection
                if (TryGetPower(GetCombatPower(CombatSkillOrder), out power))
                    return power;
            }

            Logger.Log(TrinityLogLevel.Verbose, LogCategory.Behavior, "DemonHunter GetPower() Returning DefaultPower Target={0}",
                (CurrentTarget == null) ? "Null" : CurrentTarget.InternalName);

            return DefaultPower;
        }

        /// <summary>
        /// The combat skills and the order they should be evaluated in
        /// </summary>
        private static List<Skill> CombatSkillOrder
        {
            get
            {
                return new List<Skill>
                {
                    //Buffs
                    Skills.DemonHunter.Vengeance,
                    Skills.DemonHunter.ShadowPower,
                    Skills.DemonHunter.SmokeScreen,
                    Skills.DemonHunter.Preparation,
                    Skills.DemonHunter.Companion,

                    // Cooldown only
                    Skills.DemonHunter.RainOfVengeance,

                    // Spenders
                    Skills.DemonHunter.Sentry,
                    Skills.DemonHunter.Caltrops,
                    Skills.DemonHunter.MarkedForDeath,
                    Skills.DemonHunter.FanOfKnives,
                    Skills.DemonHunter.Strafe,
                    Skills.DemonHunter.SpikeTrap,
                    Skills.DemonHunter.ClusterArrow,
                    Skills.DemonHunter.Multishot,
                    Skills.DemonHunter.RapidFire,
                    Skills.DemonHunter.Impale,
                    Skills.DemonHunter.Chakram,
                    Skills.DemonHunter.ElementalArrow,

                    // Generators
                    Skills.DemonHunter.EvasiveFire,
                    Skills.DemonHunter.HungeringArrow,
                    Skills.DemonHunter.EntanglingShot,
                    Skills.DemonHunter.Bolas,
                    Skills.DemonHunter.Grenade,
                };
            }
        }

        /// <summary>
        /// When skills should be cast, evaluated by CanCast() calls.
        /// </summary>
        public static void SetConditions()
        {
            Skills.DemonHunter.Vengeance.Meta.CastCondition = VengeanceCondition;
            Skills.DemonHunter.ShadowPower.Meta.CastCondition = ShadowPowerCondition;
            Skills.DemonHunter.SmokeScreen.Meta.CastCondition = SmokeScreenCondition;
            Skills.DemonHunter.Preparation.Meta.CastCondition = PreperationCondition;
            Skills.DemonHunter.RainOfVengeance.Meta.CastCondition = RainOfVengeanceCondition;
            Skills.DemonHunter.Sentry.Meta.CastCondition = SentryCondition;
            Skills.DemonHunter.Caltrops.Meta.CastCondition = CaltropsCondition;
            Skills.DemonHunter.Companion.Meta.CastCondition = CompanionCondition;
            Skills.DemonHunter.MarkedForDeath.Meta.CastCondition = MarkedForDeathCondition;
            Skills.DemonHunter.FanOfKnives.Meta.CastCondition = FanOfKnivesCondition;
            Skills.DemonHunter.ClusterArrow.Meta.CastCondition = ClusterArrowCondition;
            Skills.DemonHunter.Chakram.Meta.CastCondition = ChakramCondition;
            Skills.DemonHunter.Multishot.Meta.CastCondition = MultiShotCondition;
            Skills.DemonHunter.ElementalArrow.Meta.CastCondition = ElementalArrowCondition;
            Skills.DemonHunter.RapidFire.Meta.CastCondition = RapidFireCondition;
            Skills.DemonHunter.Impale.Meta.CastCondition = ImpaleCondition;
            Skills.DemonHunter.Strafe.Meta.CastCondition = StrafeCondition;
            Skills.DemonHunter.SpikeTrap.Meta.CastCondition = SpikeTrapCondition;
            Skills.DemonHunter.EvasiveFire.Meta.CastCondition = EvasiveFireCondition;
            Skills.DemonHunter.HungeringArrow.Meta.CastCondition = HungeringArrowCondition;
            Skills.DemonHunter.EntanglingShot.Meta.CastCondition = EntanglingShotCondition;
            Skills.DemonHunter.Bolas.Meta.CastCondition = BolasCondition;
            Skills.DemonHunter.Grenade.Meta.CastCondition = GrenadeCondition;
        }

        /// <summary>
        /// When Grenade should be cast
        /// </summary>
        private static bool GrenadeCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (!UseOOCPower)
                meta.TargetSelector = ret => CurrentTarget.Position;

            return true;
        }

        /// <summary>
        /// When Bolas should be cast
        /// </summary>
        private static bool BolasCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (!UseOOCPower)
                meta.TargetSelector = ret => CurrentTarget.Position;

            return true;
        }

        /// <summary>
        /// When Entangling Shot should be cast
        /// </summary>
        private static bool EntanglingShotCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (!UseOOCPower)
                meta.TargetSelector = ret => CurrentTarget.Position;

            return true;
        }

        /// <summary>
        /// When Hungering Arrow should be cast
        /// </summary>
        private static bool HungeringArrowCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (!UseOOCPower)
                meta.TargetSelector = ret => CurrentTarget.Position;

            return true;
        }

        /// <summary>
        /// When Evasive fire should be cast
        /// </summary>
        /// <param name="meta"></param>
        /// <returns></returns>
        private static bool EvasiveFireCondition(SkillMeta meta)
        {
            meta.CastRange = 14f;

            if (UseCombatPower && Sets.EmbodimentOfTheMarauder.IsMaxBonusActive && !ShouldRefreshBastiansGeneratorBuff)
                meta.CastRange = RangedAttackRange;

            if (!UseOOCPower)
                meta.TargetSelector = ret => CurrentTarget.Position;

            return true;
        }

        /// <summary>
        /// When Impale should be cast
        /// </summary>
        private static bool ImpaleCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            // Not enough resource
            if (Player.PrimaryResource <= EnergyReserve)
                return false;

            if (!TargetUtil.AnyMobsInRange(12, 4) && CurrentTarget.RadiusDistance <= 75f)
                return true;

            return false;
        }

        /// <summary>
        /// When Rapid Fire should be cast
        /// </summary>
        /// <param name="meta"></param>
        /// <returns></returns>
        private static bool RapidFireCondition(SkillMeta meta)
        {
            meta.CastFlags = CanCastFlags.NoTimer;
            meta.CastRange = RangedAttackRange;

            if (IsWaitingForSpecial)
                return false;

            // Stay above minimum resource level
            if (Player.PrimaryResource < EnergyReserve || Player.PrimaryResource < DemonHunterSettings.RapidFireMinHatred)
                return false;

            // Never use it twice in a row
            if (LastPowerUsed == SNOPower.DemonHunter_RapidFire)
                return false;

            return true;
        }

        /// <summary>
        /// When Chakram should be cast.
        /// </summary>
        private static bool ChakramCondition(SkillMeta meta)
        {
            meta.CastRange = Math.Min(50f, RangedAttackRange);

            // Spam it for Shuriken Cloud buff
            if (Runes.DemonHunter.ShurikenCloud.IsActive && TimeSincePowerUse(SNOPower.DemonHunter_Chakram) >= 110000 &&
                ((Player.PrimaryResource >= 10 && !IsWaitingForSpecial) || Player.PrimaryResource >= MinEnergyReserve))
                return true;

            // Always cast with Spines of Seething Hatred rune, grants 4 hatred
            if (Legendary.SpinesOfSeethingHatred.IsEquipped)
                return true;

            // Monsters nearby
            if (TargetUtil.ClusterExists(45f,4))
                return true;

            return false;
        }

        /// <summary>
        /// When Cluster Arrow should be cast
        /// </summary>
        private static bool ClusterArrowCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            // Natalyas - Wait for damage buff
            if (Sets.NatalyasVengeance.IsFullyEquipped && Player.PrimaryResource < 100 && !CacheData.Buffs.HasBuff(SNOPower.P2_ItemPassive_Unique_Ring_053))
                return false;

            // Stay above minimum resource level
            if (Player.PrimaryResource < EnergyReserve)
                return false;

            return true;
        }

        /// <summary>
        /// When Elemental Arrow should be cast
        /// </summary>
        private static bool ElementalArrowCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (IsWaitingForSpecial)
                return false;

            // Stay above minimum resource level
            if (Player.PrimaryResource < EnergyReserve)
                return false;

            return true;
        }

        /// <summary>
        /// When spike trap should be cast
        /// </summary>
        private static bool SpikeTrapCondition(SkillMeta meta)
        {
            meta.TargetSelector = SpikeTrapTargetSelector;

            if (LastPowerUsed != SNOPower.DemonHunter_SpikeTrap)
                return true;

            return false;
        }

        /// <summary>
        /// Where Spike trap should be cast
        /// </summary>
        private static Vector3? SpikeTrapTargetSelector(SkillMeta skillMeta)
        {
            // For distant monsters, try to target a little bit in-front of them (as they run towards us), if it's not a treasure goblin
            float reducedDistance = 0f;
            if (CurrentTarget.Distance > 17f && !CurrentTarget.IsTreasureGoblin)
            {
                reducedDistance = CurrentTarget.Distance - 17f;
                if (reducedDistance > 5f)
                    reducedDistance = 5f;
            }
            return MathEx.CalculatePointFrom(CurrentTarget.Position, Player.Position, CurrentTarget.Distance - reducedDistance);
        }

        /// <summary>
        /// When Strafe should be cast
        /// </summary>
        private static bool StrafeCondition(SkillMeta meta)
        {
            var target = TargetUtil.GetBestZigZagPoint(10f);

            meta.CastRange = 15f;
            meta.TargetSelector = ret => target;
            meta.CastFlags = CanCastFlags.NoTimer;
            meta.RequiredResource = DemonHunterSettings.StrafeMinHatred;

            if (!Player.IsRooted && target != null)
                return true;

            return false;
        }

        /// <summary>
        /// When Multishot should be cast
        /// </summary>
        private static bool MultiShotCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;
            meta.CastFlags = CanCastFlags.NoPowerManager;
            meta.CastFlags = CanCastFlags.NoTimer;

            if (IsWaitingForSpecial)
                return false;

            // Stay above minimum resource level
            if (Player.PrimaryResource < EnergyReserve)
                return false;

            // Natalyas - Wait for damage buff
            if (Sets.NatalyasVengeance.IsFullyEquipped && Player.PrimaryResource < 100 && !CacheData.Buffs.HasBuff(SNOPower.P2_ItemPassive_Unique_Ring_053))
                return false;

            return true;
        }

        /// <summary>
        /// When Fan of Knives should be cast
        /// </summary>
        private static bool FanOfKnivesCondition(SkillMeta meta)
        {
            meta.CastRange = 25f;

            if (TargetUtil.EliteOrTrashInRange(15) || TargetUtil.AnyTrashInRange(15f, 5, false))
                return true;

            return false;
        }

        /// <summary>
        /// When Marked for Death should be cast
        /// </summary>
        private static bool MarkedForDeathCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;
            meta.CastFlags = CanCastFlags.NoTimer;

            if (!CurrentTarget.HasDebuff(SNOPower.DemonHunter_MarkedForDeath) && !SpellTracker.IsUnitTracked(CurrentTarget, SNOPower.DemonHunter_MarkedForDeath))
                return true;

            return false;
        }

        /// <summary>
        /// When Companion should be cast
        /// </summary>
        private static bool CompanionCondition(SkillMeta meta)
        {
            meta.CastFlags = CanCastFlags.NoTimer;

            // Use Spider Slow on 4 or more trash mobs in an area or on Unique/Elite/Champion
            if (Runes.DemonHunter.SpiderCompanion.IsActive && TargetUtil.ClusterExists(25f, 4) && TargetUtil.EliteOrTrashInRange(25f))
                return true;

            //Use Bat when Hatred is Needed
            if (Runes.DemonHunter.BatCompanion.IsActive && Player.PrimaryResourceMissing >= 60)
                return true;

            // Use Boar Taunt on 3 or more trash mobs in an area or on Unique/Elite/Champion
            if (Runes.DemonHunter.BoarCompanion.IsActive && ((TargetUtil.ClusterExists(20f, 4) && TargetUtil.EliteOrTrashInRange(20f)) || (CurrentTarget.IsBossOrEliteRareUnique && CurrentTarget.Distance <= 20f)))
                return true;

            // Ferrets used for picking up Health Globes when low on Health
            if (Runes.DemonHunter.FerretCompanion.IsActive && Trinity.ObjectCache.Any(o => o.Type == GObjectType.HealthGlobe && o.Distance < 60f) && Player.CurrentHealthPct < EmergencyHealthPotionLimit)
                return true;

            // Use Wolf Howl on Unique/Elite/Champion - Would help for farming trash, but trash farming should not need this - Used on Elites to reduce Deaths per hour
            if (Runes.DemonHunter.WolfCompanion.IsActive && (TargetUtil.AnyElitesInRange(100f) || TargetUtil.AnyMobsInRange(40, 8)))
                return true;

            // Companion off CD
            if (DemonHunterSettings.CompanionOffCooldown && TargetUtil.AnyMobsInRange(RangedAttackRange))
                return true;

            return false;
        }

        /// <summary>
        /// When Caltrops should be cast
        /// </summary>
        private static bool CaltropsCondition(SkillMeta meta)
        {
            return TargetUtil.AnyMobsInRange(40) && !GetHasBuff(SNOPower.DemonHunter_Caltrops);
        }

        /// <summary>
        /// When Sentry should be cast
        /// </summary>
        /// <param name="meta"></param>
        /// <returns></returns>
        private static bool SentryCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;
            meta.CastFlags = CanCastFlags.NoTimer;

            if (IsZeisOfStoneEquipped && !Runes.DemonHunter.PolarStation.IsActive)
                meta.TargetSelector = SentryTargetSelector;

            if (TargetUtil.AnyMobsInRange(65) && (Trinity.PlayerOwnedDHSentryCount < MaxSentryCount || !IsSentryOnTarget(2, 30f)))
                return true;

            return false;
        }

        /// <summary>
        /// Where Spike trap should be cast
        /// </summary>
        private static Vector3? SentryTargetSelector(SkillMeta skillMeta)
        {
            Vector3 offSetTarget = MathEx.CalculatePointFrom(Player.Position, CurrentTarget.Position, 51f);
            if (offSetTarget.Distance2D(LastZeiOfStoneLocation) >= 25f && CurrentTarget.IsInLineOfSight)
            {
                LastZeiOfStoneLocation = offSetTarget;
                return LastZeiOfStoneLocation;
            }

            return null;
        }

        /// <summary>
        /// When Rain of Vengeance should be cast
        /// </summary>
        private static bool RainOfVengeanceCondition(SkillMeta meta)
        {
            meta.CastRange = RangedAttackRange;

            if (DemonHunterSettings.RainOfVengeanceOffCD || Sets.NatalyasVengeance.IsFirstBonusActive)
            {
                meta.CastFlags = CanCastFlags.NoTimer;
                return true;
            }

            if (TargetUtil.ClusterExists(45f, 4) || TargetUtil.AnyElitesInRange(90f))
                return true;

            return false;
        }

        /// <summary>
        /// When Preperation should be cast
        /// </summary>
        private static bool PreperationCondition(SkillMeta meta)
        {
            meta.AfterUseDelay = Runes.DemonHunter.FocusedMind.IsActive ? 15000 : 500;
            meta.CastFlags = CanCastFlags.NoTimer;

            if (!Runes.DemonHunter.Punishment.IsActive && Player.SecondaryResource <= V.F("DemonHunter.MinPreparationDiscipline"))
                return true;

            if (Runes.DemonHunter.Punishment.IsActive && Player.PrimaryResourceMissing > 75)
                return true;

            return false;
        }

        /// <summary>
        /// When Smoke Screen should be cast
        /// </summary>
        private static bool SmokeScreenCondition(SkillMeta meta)
        {
            meta.CastFlags = CanCastFlags.NoTimer;

            if (Player.StandingInAvoidance)
                return true;

            // Buff Already Active
            if (meta.Skill.IsBuffActive)
                return false;

            // Mobs in range
            if (TargetUtil.AnyMobsInRange(15) || (Legendary.MeticulousBolts.IsEquipped && TargetUtil.AnyMobsInRange(60)))
                return true;

            // Defensive Cast
            if((Player.CurrentHealthPct <= 0.50 || Player.IsRooted || Player.IsIncapacitated))
                return true;

            if (!Player.IsMoving && Trinity.IsTryingToTownPortal)
                return false;

            // Spam Setting
            if (DemonHunterSettings.SpamSmokeScreen)
                return true;

            return false;
        }

        /// <summary>
        /// When Shadow Power should be cast
        /// </summary>
        private static bool ShadowPowerCondition(SkillMeta meta)
        {
            // Buff Already Active
            if(GetHasBuff(SNOPower.DemonHunter_ShadowPower))
                return false;

            // Not Enough Discipline
            if (Player.SecondaryResource >= 14)
                return false;

            // Used Recently
            if (TimeSincePowerUse(SNOPower.DemonHunter_ShadowPower) < 4500)
                return false;

            // Low Health
            if(Player.CurrentHealthPct <= CombatBase.EmergencyHealthPotionLimit && Player.SecondaryResource >= 14)
                return true;

            // Defensive Cast
            if(Player.IsRooted || TargetUtil.AnyMobsInRange(15))
                return true;

            // Spam Setting
            if(DemonHunterSettings.SpamShadowPower)
                return true;

            return false;
        }

        /// <summary>
        /// When Vengeance should be cast
        /// </summary>
        private static bool VengeanceCondition(SkillMeta meta)
        {
            meta.CastFlags = CanCastFlags.NoTimer;

            if (!DemonHunterSettings.VengeanceElitesOnly && TargetUtil.AnyMobsInRange(60f, 6))
                return true;

            if (TargetUtil.IsEliteTargetInRange(100f))
                return true;

            return false;
        }
    }
}
