﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Trinity.Combat;
using Trinity.Combat.Abilities;
using Trinity.DbProvider;
using Trinity.Helpers;
using Trinity.Technicals;
using Zeta.Bot;
using Zeta.Bot.Navigation;
using Zeta.Game;
using Zeta.Game.Internals.Actors;
using Logger = Trinity.Technicals.Logger;

namespace Trinity
{
    public partial class Trinity
    {
        internal static int LastSceneId = -1;

        internal static Stopwatch HotbarRefreshTimer = new Stopwatch();

        private static bool TargetCheckResult(bool result, string source)
        {
            Logger.Log(TrinityLogLevel.Debug, LogCategory.GlobalHandler, "TargetCheck returning {0}, {1}", result, source);
            return result;
        }

        /// <summary>
        /// Find fresh targets, start main BehaviorTree if needed, cast any buffs needed etc.        
        /// </summary>
        /// <param name="ret"></param>
        /// <returns></returns>
        internal static bool TargetCheck(object ret)
        {
            using (new PerformanceLogger("TargetCheck"))
            {
                if (Player.IsDead)
                {
                    return TargetCheckResult(false, "Is Dead");
                }

                RefreshDiaObjects();
                Inferno.ExecuteRoutine();

                // We have a target, start the target handler!
                if (CurrentTarget != null)
                {
                    ShouldPickNewAbilities = true;
                    return TargetCheckResult(true, "Current Target is not null, Target: " + CurrentTarget.Type.ToString() + " - " + CurrentTarget.InternalName);
                }

                // if we just opened a horadric cache, wait around to open it
                if (DateTime.UtcNow.Subtract(Composites.LastFoundHoradricCache).TotalSeconds < 5)
                    return TargetCheckResult(true, "Recently opened Horadric Cache");

                using (new PerformanceLogger("TargetCheck.OOCPotion"))
                {
                    // Pop a potion when necessary
                    if (Player.CurrentHealthPct <= CombatBase.EmergencyHealthPotionLimit)
                    {
                        UsePotionIfNeededTask();
                    }
                }

                TrinityStatusText = "[Trinity] No more targets - DemonBuddy/profile management is now in control";

                if (Settings.Advanced.DebugInStatusBar && ResetTrinityStatusText)
                {
                    ResetTrinityStatusText = false;
                    BotMain.StatusText = TrinityStatusText;
                }


                using (new PerformanceLogger("TargetCheck.OOCBuff"))
                {
                    if (!WantToTownRun && !ForceVendorRunASAP)
                    {
                        BarbarianCombat.AllowSprintOOC = true;

                        TrinityPower powerOOC = AbilitySelector();
                        if (Player.IsCastingOrLoading && TargetUtil.AnyMobsInRange(35f) && 
                            (powerOOC.SNOPower == SNOPower.X1_Crusader_AkaratsChampion ||
                            powerOOC.SNOPower == SNOPower.Barbarian_WrathOfTheBerserker ||
                            powerOOC.SNOPower == SNOPower.X1_Monk_Epiphany) &&
                            CombatBase.Cast(powerOOC))
                        {
                            Logger.Log(TrinityLogLevel.Verbose, LogCategory.Behavior, "Using OOC Buff for TownPortal 0 interupt: {0}", powerOOC.SNOPower.ToString());
                        }
                        else if (!Player.IsCastingOrLoading && CombatBase.Cast(powerOOC))
                        {
                            Logger.Log(TrinityLogLevel.Verbose, LogCategory.Behavior, "Using OOC Buff: {0}", powerOOC.SNOPower.ToString());
                        }
                    }
                }

                CurrentTarget = null;

                if ((ForceVendorRunASAP || WantToTownRun) && TownRun.TownRunTimerRunning())
                {
                    Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Waiting for town run timer (Target Check)", true);
                    return TargetCheckResult(true, "Waiting for TownRunTimer");
                }

                return TargetCheckResult(false, "End of TargetCheck");
            }
        }
    }
}
