﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Trinity.Config;
using Trinity.DbProvider;
using Trinity.ItemRules;
using Trinity.Reference;
using Trinity.Technicals;
using Zeta.Bot.Navigation;
using Zeta.Common;
using Zeta.Common.Plugins;
using Zeta.Game.Internals.Actors;

namespace Trinity
{
    public partial class Trinity : IPlugin
    {
        internal static CacheData.PlayerCache Player { get { return CacheData.Player; } }
        internal static MainGridProvider MainGridProvider { get { return (MainGridProvider)Navigator.SearchGridProvider; } }

        private static TrinitySetting _settings = new TrinitySetting();
        internal static TrinitySetting Settings 
        { 
            get { return _settings; }
            set { _settings = value; }
        }

        private static bool _isPluginEnabled = false;
        internal static bool IsPluginEnabled
        {
            get { return _isPluginEnabled; }
            set { _isPluginEnabled = value; }
        }

        private static SNOPower _lastPowerUsed = SNOPower.None;
        public static SNOPower LastPowerUsed
        {
            get { return _lastPowerUsed; }
            set { _lastPowerUsed = value; }
        }

        private static bool _isWaitingForSpecial = false;
        internal static bool IsWaitingForSpecial
        {
            get { return _isWaitingForSpecial; }
            set { _isWaitingForSpecial = value; }
        }

        private static bool _wantToTownRun = false;
        public static bool WantToTownRun
        {
            get { return _wantToTownRun; }
            set { _wantToTownRun = value; }
        }

        private static int _totalDeaths = 0;
        internal static int TotalDeaths
        {
            get { return _totalDeaths; }
            set { _totalDeaths = value; }
        }

        private static Vector3 eventStartPosition = Vector3.Zero;
        public static Vector3 EventStartPosition
        {
            get { return Trinity.eventStartPosition; }
            set { Trinity.eventStartPosition = value; }
        }

        private static DateTime eventStartTime = DateTime.MinValue;
        public static DateTime EventStartTime
        {
            get { return Trinity.eventStartTime; }
            set { Trinity.eventStartTime = value; }
        }

        internal static TrinityCacheObject CurrentTarget = null;
        internal static TrinityCacheObject LastTarget = null;
        internal static List<TrinityCacheObject> ObjectCache = new List<TrinityCacheObject>();

        internal static bool ShouldPickNewAbilities = false;
        internal static bool ResetTrinityStatusText = false;
        internal static bool AnyTreasureGoblinsPresent = false;
        internal static bool AnyMobsInRange = false;
        internal static bool ShouldStayPutDuringAvoidance = false;
        public static bool ForceVendorRunASAP = false;
        internal static bool TargetIsInLoS = false;
        internal static bool TargetIsInRange = false;
        internal static bool KillMonstersInAoE = false;
        internal static bool LogHasFlagPerformance = false;
        internal static bool IsTryingToTownPortal = false;

        internal static int IgnoredRActorGuid = -1;
        internal static int IgnoredRActorGuidForLoops = 0;
        internal static int DeathsThisRun = 0;
        internal static int TotalBountyCachesOpened = 0;
        internal static int PlayerOwnedMysticAllyCount = 0;
        internal static int PlayerOwnedGargantuanCount = 0;
        internal static int PlayerOwnedZombieDogCount = 0;
        internal static int PlayerOwnedFetishCount = 0;
        internal static int PlayerOwnedDHPetsCount = 0;
        internal static int PlayerOwnedDHSentryCount = 0;
        internal static int PlayerOwnedHydraCount = 0;

        internal static double TargetLastHealth;
        internal static double MinEnergyReserve = 0d;

        internal static float KillRange = 0f;
        internal static float LootRange = 0f;
        internal static float TargetDistance = 0f;

        internal static DateTime LastPickNewAbilitiesForced = DateTime.MinValue;
        internal static DateTime LastDeathTime = DateTime.MinValue;
        internal static DateTime LastSentMovePower = DateTime.MinValue;
        internal static DateTime LastGlobalCooldownUse = DateTime.MinValue;
        internal static DateTime LastHadUnitInSights = DateTime.MinValue;
        internal static DateTime LastHadEliteUnitInSights = DateTime.MinValue;
        internal static DateTime LastHadBossUnitInSights = DateTime.MinValue;
        internal static DateTime LastHadContainerInSights = DateTime.MinValue;
        internal static DateTime Blacklist15LastClear = DateTime.MinValue;
        internal static DateTime Blacklist60LastClear = DateTime.MinValue;
        internal static DateTime Blacklist90LastClear = DateTime.MinValue;

        internal static Vector3 LastMoveToTarget = new Vector3();
        internal static Vector3 TargetPosition = new Vector3();
        internal static Vector3 MovePosition = new Vector3();

        internal static string TrinityStatusText = string.Empty;
        internal static string CurrentProfile = string.Empty;
        internal static string CurrentProfileName = string.Empty;
        internal static string FirstProfile = string.Empty;

        internal static Regex NameNumberTrimRegex = new Regex(@"-\d+$", RegexOptions.Compiled);
        internal static Interpreter StashRule = null;
        internal static Dictionary<int, Tuple<int, DateTime>> TimedBlacklist = new Dictionary<int, Tuple<int, DateTime>>();
        internal static Tuple<int, DateTime> IgnoreDuringSeconds(int seconds) { return new Tuple<int, DateTime>(seconds, DateTime.UtcNow); }

        /// <summary>
        /// Set bool value, clear blacklist, refresh skill, etc.
        /// </summary>
        internal static void ResetTickValues()
        {
            IsTryingToTownPortal = TownRun.IsTryingToTownPortal();
            LogHasFlagPerformance = Settings.Advanced.LogCategories.HasFlag(LogCategory.Performance);
            KillMonstersInAoE = Settings.Combat.Misc.KillMonstersInAoE;

            SkillUtils.ByActorClass(Player.ActorClass).ForEach(s => s.RefreshOnTick());

            TimedBlacklist.ToList()
                .Where(i => DateTime.UtcNow.Subtract(i.Value.Item2).TotalSeconds >= i.Value.Item1).ToList()
                .ForEach(i => TimedBlacklist.Remove(i.Key));
        }
    }
}
