﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Markup;
using System.Windows.Media;
using Trinity.Combat;
using Trinity.Combat.Abilities;
using Trinity.Technicals;
using Zeta.Common;
using Zeta.Common.Xml;
using Zeta.Game;
using Zeta.XmlEngine;
using Logger = Trinity.Technicals.Logger;

namespace Trinity.Helpers
{
    public class Warden
    {
        public static bool IsExist = false;
        public static bool IsDead = false;
        public TrinityCacheObject TrinityObject = null;

        public Warden()
        {
            TrinityCacheObject obj = null;
            try
            {
                obj = Trinity.ObjectCache.First(o => DataDictionary.KeyWardenIds.Contains(o.ActorSNO));
            }
            catch { }

            if (obj == null)
            {
                DeclareDead();
                return;
            }

            DeclareFound();

            TrinityObject = obj;
            Inferno.LastWardenScanTimed = DateTime.Now;
        }

        public bool IsInRange(float range)
        {
            return TrinityObject != null && TrinityObject.Distance <= range;
        }

        public bool HitPointBelow(float hitPoint)
        {
            return TrinityObject != null && TrinityObject.HitPointsPct < hitPoint;
        }

        public static void DeclareDead()
        {
            if (!IsDead && IsExist)
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "[Inferno] KeyWarden vanquished");
                Inferno.IncrementCount("Dead");

                IsDead = true;
                IsExist = false;
            }
        }

        public static void DeclareFound()
        {
            if (!IsExist && !IsDead)
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "[Inferno] KeyWarden found");
                Inferno.IncrementCount("Found");

                IsDead = false;
                IsExist = true;
            }
        }

        public static void Reset()
        {
            IsExist = false;
            IsDead = false;
        }
    }

    public class Inferno
    {
        public static readonly List<int> KeySNOs = new List<int> { 364694, 364695, 364696, 364697 };

        public static DateTime RunTime = DateTime.MinValue;

        public static int CurrentAct = 0;
        public static string RunStatus = "None";

        public static List<int> RunByActCount = new List<int> { 0, 0, 0, 0 };
        public static List<int> DefeatedWardensCount = new List<int> { 0, 0, 0, 0 };
        public static List<int> FoundWardensCount = new List<int> { 0, 0, 0, 0 };
        public static List<int> InitCount = new List<int> { -1, -1, -1, -1 };
        public static List<int> KeyCount = new List<int> { 0, 0, 0, 0 };
        public static List<int> TotalDropCount = new List<int> { 0, 0, 0, 0 };

        public static HashSet<int> UsedAct = new HashSet<int>();

        public static DateTime LastWardenScanTimed = DateTime.MinValue;
        public static bool WaitProbablyDeadIsFinished 
        { 
            get { return DateTime.Now.Subtract(LastWardenScanTimed).TotalSeconds > 2.5; } 
        }

        public static int GameCount 
        { 
            get { return Trinity.TotalLeaveGames; } 
        }

        public static void ExecuteRoutine()
        {
            if (!Trinity.Player.IsInInfernoRun)
            {
                Warden.DeclareDead();
                Warden.Reset();
                return;
            }

            InfernoUI.Refresh();

            if (Trinity.ObjectCache == null || Trinity.ObjectCache.Count() <= 0)
            {
                Warden.DeclareDead();
                Warden.Reset();
                return;
            }

            Warden warden = new Warden();
            if (Warden.IsExist && !Warden.IsDead)
            {
                if (Trinity.ObjectCache.Any(o => KeySNOs.Contains(o.ActorSNO)))
                {
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "[Inferno] Dead.KeyOnFloor");
                    Inferno.IncrementCount("Key");
                    Warden.DeclareDead();
                }
                else if (warden.IsInRange(70f) && WaitProbablyDeadIsFinished)
                {
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "[Inferno] Dead.InRange");
                    Warden.DeclareDead();
                }
                else if (warden.HitPointBelow(0.03f) && !IsRunOnAct(3))
                {
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "[Inferno] Dead.LowHp");
                    Warden.DeclareDead();
                }
                else if (warden.HitPointBelow(0.05f) && warden.IsInRange(50f) && !IsRunOnAct(3) && !IsRunOnAct(2))
                {
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "[Inferno] Dead.InRangeLowHp");
                    Warden.DeclareDead();
                }
                else if (!warden.IsInRange(45f))
                {
                    CombatBase.QueuedMovement.Queue(new QueuedMovement
                    {
                        Id = warden.TrinityObject.ActorSNO,
                        Name = warden.TrinityObject.InternalName,
                        Infos = "(MoveToWarden) " + warden.TrinityObject.Infos,
                        Destination = warden.TrinityObject.Position,
                        StopCondition = m =>
                            m.Destination == Vector3.Zero ||
                            m.Destination.Distance2D(Trinity.Player.Position) <= 1f ||
                            warden == null ||
                            warden.TrinityObject == null ||
                            warden.IsInRange(44f) ||
                            Warden.IsDead
                        ,
                        Options = new QueuedMovementOptions
                        {
                            Logging = LogLevel.Info,
                            AcceptableDistance = 1f,
                            Type = MoveType.SpecialCombat,
                        }
                    });
                }
            }
        }

		public static bool IsRunOnAct(int act)
		{
			return CurrentAct == act;
		}

        public static void IncrementCount(string counter, string counter2 = null)
        {
            Increment(counter);

            if (counter2 != null)
                Increment(counter2);
        }

        public static void RefreshKeyCount()
        {
            if (!Trinity.Player.IsFullyValid || CacheData.Inventory.Backpack == null || CacheData.Inventory.Backpack.Count() <= 0)
                return;

            for (int i = 0; i <= 3; i++)
            {
                try
                {
                    KeyCount[i] = CacheData.Inventory.Backpack.Where(o => o != null && o.ActorSNO == KeySNOs[i]).Sum(o => o.ItemStackQuantity) +
                        CacheData.Inventory.Stash.Where(o => o != null && o.ActorSNO == KeySNOs[i]).Sum(o => o.ItemStackQuantity);

                    if (InitCount[i] < 0)
                        InitCount[i] = KeyCount[i];

                    if (InitCount[i] < KeyCount[i])
                        TotalDropCount[i] = KeyCount[i] - InitCount[i];
                }
                catch { }

                
            }
        }

        public static void Increment(string counter)
        {
            if (CurrentAct <= 0)
                return;

            switch (counter)
            {
                case "Key":
                    TotalDropCount[CurrentAct - 1]++;
                    break;
                case "Found":
                    if (FoundWardensCount[CurrentAct - 1] < RunByActCount[CurrentAct - 1])
                        FoundWardensCount[CurrentAct - 1]++;
                    break;
                case "Profile":
                    RunByActCount[CurrentAct - 1]++;
                    break;
                case "Dead":
                    if (DefeatedWardensCount[CurrentAct - 1] < RunByActCount[CurrentAct - 1])
                        DefeatedWardensCount[CurrentAct - 1]++;
                    break;
                default:
                    break;
            }
        }

        public static bool AvailableToRun(int act)
        {
            if (act < 1)
                return false;

            RefreshKeyCount();

            if (UsedAct.Contains(act))
                return false;

            int actMaxKeys = KeyCount.IndexOf(KeyCount.Max()) + 1;
            int actMinKeys = KeyCount.IndexOf(KeyCount.Min()) + 1;

            if (KeyCount.Min() == KeyCount.Max())
                return true;

            if (KeyCount[act - 1] == KeyCount[actMaxKeys - 1] && KeyCount[actMaxKeys - 1] > 0 && KeyCount[act - 1] - 3 <= KeyCount[actMinKeys - 1])
                return true;

            if (KeyCount[act - 1] != KeyCount[actMaxKeys - 1])
                return true;

            return false;
        }

        #region RunStatus.Debug
        public static void DebugRunStats()
        {
            double time = DateTime.Now.Subtract(RunTime).TotalSeconds / 3600;

            double kh = TotalDropCount.Sum() / time;
            string skh = kh.ToString("F2");
            string s = TotalDropCount.Sum().ToString("D3");
            string s1 = GameCount.ToString("D4");
            string s3 = DefeatedWardensCount.Sum().ToString("D4");
            double d = (double)TotalDropCount.Sum() / GameCount;
            string s2 = d.ToString("F3");
            double d2 = (double)TotalDropCount.Sum() / DefeatedWardensCount.Sum();
            string s5 = d2.ToString("F3");
            double d3 = (double)TotalDropCount.Sum() / RunByActCount.Sum();
            string s6 = d3.ToString("F3");
            double d4 = (double)DefeatedWardensCount.Sum() / RunByActCount.Sum();
            string s7 = d4.ToString("F3");
            TimeSpan RunTimed = DateTime.Now - RunTime;

            Logger.LogDebug("¤ Run Timer: " + RunTimed.ToString(@"hh\:mm\:ss"));
            Logger.LogDebug("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤");
            Logger.LogDebug("¤                       ~  INFERNAL KEY STATS  ~                         ¤");

            if (TotalDropCount.Sum() > 0)
                Logger.LogDebug("¤                              TOTAL: " + s + " KEYS !                                ¤");
            Logger.LogDebug("¤                         KEYS/HOUR:   " + skh + "   K/H                            ¤");
            if (GameCount > 0)
                Logger.LogDebug("¤                        Completed GAMES: " + s1 + "                           ¤");
            if (GameCount > 0 || RunByActCount.Sum() > 0 || DefeatedWardensCount.Sum() > 0)
                Logger.LogDebug("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤");
            if (DefeatedWardensCount.Sum() > 0)
                Logger.LogDebug("¤         Warden vanquished:           " + s3 + " w                           ¤");
            if (RunByActCount.Sum() > 0)
                Logger.LogDebug("¤         Warden per profile:            " + s7 + " w/p                       ¤");
            if (GameCount > 0)
                Logger.LogDebug("¤         Keys per game:                  " + s2 + " k/g                        ¤");
            if (RunByActCount.Sum() > 0)
                Logger.LogDebug("¤         Keys per profile:                 " + s6 + " k/p                        ¤");
            if (DefeatedWardensCount.Sum() > 0)
                Logger.LogDebug("¤         Keys per Warden:               " + s5 + " k/w                       ¤");

            string s8 = KeyCount[0].ToString("D3");
            string s9 = KeyCount[1].ToString("D3");
            string s10 = KeyCount[2].ToString("D3");
            string s11 = KeyCount[3].ToString("D3");
            Logger.LogDebug("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤");
            Logger.LogDebug("¤                Key counter               |      Game stats                 ¤");

            if (AvailableToRun(1))
                Logger.LogDebug(String.Format("¤  Act1:      {0}  Infernal keys      |      Available profile         ¤", s8));
            else if (UsedAct.Contains(1))
                Logger.LogDebug(String.Format("¤  Act1:      {0}  Infernal keys      |      Profile used                ¤", s8));
            else
                Logger.LogDebug(String.Format("¤  Act1:      {0}  Infernal keys      |      Not available profile  ¤", s8));

            if (AvailableToRun(2))
                Logger.LogDebug(String.Format("¤  Act2:      {0}  Infernal keys      |      Available profile         ¤", s9));
            else if (UsedAct.Contains(2))
                Logger.LogDebug(String.Format("¤  Act2:      {0}  Infernal keys      |      Profile used                ¤", s9));
            else
                Logger.LogDebug(String.Format("¤  Act2:      {0}  Infernal keys      |      Not available profile  ¤", s9));

            if (AvailableToRun(3))
                Logger.LogDebug(String.Format("¤  Act3:      {0}  Infernal keys      |      Available profile         ¤", s10));
            else if (UsedAct.Contains(3))
                Logger.LogDebug(String.Format("¤  Act3:      {0}  Infernal keys      |      Profile used                ¤", s10));
            else
                Logger.LogDebug(String.Format("¤  Act3:      {0}  Infernal keys      |      Not available profile  ¤", s10));

            if (AvailableToRun(4))
                Logger.LogDebug(String.Format("¤  Act4:      {0}  Infernal keys      |      Available profile         ¤", s11));
            else if (UsedAct.Contains(4))
                Logger.LogDebug(String.Format("¤  Act4:      {0}  Infernal keys      |      Profile used                ¤", s11));
            else
                Logger.LogDebug(String.Format("¤  Act4:      {0}  Infernal keys      |      Not available profile  ¤", s11));

            Logger.LogDebug("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤");
        } 
        #endregion
    }    
}
