﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;
using Trinity.Combat;
using Trinity.Combat.Abilities;
using Trinity.Config.Combat;
using Trinity.Reference;
using Trinity.Technicals;
using Zeta.Bot;
using Zeta.Bot.Navigation;
using Zeta.Common;
using Zeta.Game;
using Zeta.Game.Internals;
using Zeta.Game.Internals.Actors;
using Logger = Trinity.Technicals.Logger;

namespace Trinity.DbProvider
{
    // Player Mover Class
    public class PlayerMover : IPlayerMover
    {
        private static readonly HashSet<int> BasicMovementOnlyIDs = new HashSet<int> { 138989, 176074, 176076, 176077, 176536, 260330, 330695, 330696, 330697, 330698, 330699 };
        // 138989 = health pool, 176074 = protection, 176076 = fortune, 176077 = frenzied, 176536 = portal in leorics, 260330 = cooldown shrine, 330695 to 330699 = pylons
        // Exp shrines = ???? Other shrines ????


        private static bool ShrinesInArea(Vector3 targetpos)
        {
            return Trinity.ObjectCache.Any(o => BasicMovementOnlyIDs.Contains(o.ActorSNO) && Vector3.Distance(o.Position, targetpos) <= 50f);
        }

        private static readonly DateTime LastUsedMoveStop = DateTime.MinValue;
        public void MoveStop()
        {
            if (DateTime.UtcNow.Subtract(LastUsedMoveStop).TotalMilliseconds < 250)
                return;

            ZetaDia.Me.UsePower(SNOPower.Walk, ZetaDia.Me.Position, Trinity.Player.WorldDynamicID);
        }

        // Anti-stuck variables
        internal static Vector3 LastMoveToTarget = Vector3.Zero;
        internal static int TimesReachedStuckPoint = 0;
        internal static int TotalAntiStuckAttempts = 1;
        internal static Vector3 vSafeMovementLocation = Vector3.Zero;
        internal static DateTime TimeLastRecordedPosition = DateTime.MinValue;
        internal static Vector3 LastPosition = Vector3.Zero;
        internal static DateTime LastGeneratedStuckPosition = DateTime.MinValue;
        internal static int TimesReachedMaxUnstucks = 0;
        internal static DateTime LastCancelledUnstucker = DateTime.MinValue;
        internal static DateTime LastRecordedAnyStuck = DateTime.MinValue;
        internal static int CancelUnstuckerForSeconds = 60;
        internal static DateTime LastRestartedGame = DateTime.MinValue;
        internal static bool UnStuckCheckerLastResult = false;
        internal static DateTime TimeLastUsedPlayerMover = DateTime.MinValue;

        internal static Vector3 LastTempestRushPosition = Vector3.Zero;

        // Store player current position
        public static Vector3 MyPosition { get { return Trinity.Player.Position; } }

        //For Tempest Rush Monks
        private static bool CanChannelTempestRush;


        public static Vector3 UnstuckHandler()
        {
            return UnstuckHandler(MyPosition, LastMoveToTarget);
        }
        // Actually deal with a stuck - find an unstuck point etc.
        public static Vector3 UnstuckHandler(Vector3 vMyCurrentPosition, Vector3 vOriginalDestination)
        {
            if (Trinity.Settings.Advanced.DisableAllMovement)
                return Vector3.Zero;

            // Update the last time we generated a path
            LastGeneratedStuckPosition = DateTime.UtcNow;

            Trinity.CurrentTarget = TargetUtil.GetClosestTarget(25f, useWeights: false);
            if (CurrentTarget != null && CurrentTarget.IsUnit)
            {
                CombatBase.Cast(Trinity.AbilitySelector());
            }
            else
            {
                Trinity.CurrentTarget = TargetUtil.GetClosestDestructible(25f);
                if (CurrentTarget != null && CurrentTarget.IsDestroyable)
                {
                    CombatBase.Cast(Trinity.AbilitySelector());
                }
            }

            // If we got stuck on a 2nd/3rd/4th "chained" anti-stuck route, then return the old move to target to keep movement of some kind going
            if (TimesReachedStuckPoint > 0)
            {
                vSafeMovementLocation = Vector3.Zero;

                // Reset the path and allow a whole "New" unstuck generation next cycle
                TimesReachedStuckPoint = 0;
                // And cancel unstucking for 9 seconds so DB can try to navigate
                CancelUnstuckerForSeconds = (9 * TotalAntiStuckAttempts);
                if (CancelUnstuckerForSeconds < 20)
                    CancelUnstuckerForSeconds = 20;
                LastCancelledUnstucker = DateTime.UtcNow;
                Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "Clearing old route and trying new path find to: " + LastMoveToTarget.ToString());
                NavigateTo(LastMoveToTarget, "original destination");
                return vSafeMovementLocation;
            }
            // Only try an unstuck 10 times maximum in XXX period of time
            if (Vector3.Distance(vOriginalDestination, vMyCurrentPosition) >= V.F("Unstucker.MaxDistance"))
            {
                Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "You are " + Vector3.Distance(vOriginalDestination, vMyCurrentPosition).ToString() + " distance away from your destination.");
                Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "This is too far for the unstucker, and is likely a sign of ending up in the wrong map zone.");
                TotalAntiStuckAttempts = 20;
            }

            if (TotalAntiStuckAttempts <= 10)
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.Movement, "Your bot got stuck! Trying to unstuck (attempt #{0} of 10 attempts) {1} {2} {3} {4}",
                    TotalAntiStuckAttempts.ToString(),
                    "Act=\"" + ZetaDia.CurrentAct + "\"",
                    "questId=\"" + ZetaDia.CurrentQuest.QuestSNO + "\"",
                    "stepId=\"" + ZetaDia.CurrentQuest.StepId + "\"",
                    "worldId=\"" + ZetaDia.CurrentWorldId + "\""
                );

                Logger.Log(TrinityLogLevel.Verbose, LogCategory.UserInformation, "(destination=" + vOriginalDestination.ToString() + ", which is " + Vector3.Distance(vOriginalDestination, vMyCurrentPosition).ToString() + " distance away)");

                /*
                 * Unstucker position
                 */
                //vSafeMovementLocation = NavHelper.FindSafeZone(true, TotalAntiStuckAttempts, vMyCurrentPosition);
                vSafeMovementLocation = NavHelper.SimpleUnstucker();

                // Temporarily log stuff
                if (TotalAntiStuckAttempts == 1 && Trinity.Settings.Advanced.LogStuckLocation)
                {
                    FileStream LogStream = File.Open(Path.Combine(FileManager.LoggingPath, "Stucks - " + Trinity.Player.ActorClass.ToString() + ".log"), FileMode.Append, FileAccess.Write, FileShare.Read);
                    using (StreamWriter LogWriter = new StreamWriter(LogStream))
                    {
                        LogWriter.WriteLine(DateTime.UtcNow.ToString() + ": Original Destination=" + LastMoveToTarget.ToString() + ". Current player position when stuck=" + vMyCurrentPosition.ToString());
                        LogWriter.WriteLine("Profile Name=" + ProfileManager.CurrentProfile.Name);
                    }
                    LogStream.Close();
                }
                // Now count up our stuck attempt generations
                TotalAntiStuckAttempts++;
                return vSafeMovementLocation;
            }

            TimesReachedMaxUnstucks++;
            TotalAntiStuckAttempts = 1;
            vSafeMovementLocation = Vector3.Zero;
            LastPosition = Vector3.Zero;
            TimesReachedStuckPoint = 0;
            TimeLastRecordedPosition = DateTime.MinValue;
            LastGeneratedStuckPosition = DateTime.MinValue;
            // int iSafetyLoops = 0;
            if (TimesReachedMaxUnstucks == 1)
            {
                Navigator.Clear();
                Logger.Log(TrinityLogLevel.Info, LogCategory.Movement, "Anti-stuck measures now attempting to kickstart DB's path-finder into action.");
                Navigator.MoveTo(vOriginalDestination, "original destination", false);
                CancelUnstuckerForSeconds = 40;
                LastCancelledUnstucker = DateTime.UtcNow;
                return vSafeMovementLocation;
            }
            if (TimesReachedMaxUnstucks == 2)
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.Movement, "Anti-stuck measures failed. Now attempting to reload current profile.");

                Navigator.Clear();

                ProfileManager.Load(Zeta.Bot.ProfileManager.CurrentProfile.Path);
                Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Anti-stuck successfully reloaded current profile, DemonBuddy now navigating again.");
                return vSafeMovementLocation;

                // Didn't make it to town, so skip instantly to the exit game system
                //iTimesReachedMaxUnstucks = 3;
            }
            // Exit the game and reload the profile
            if (Trinity.Settings.Advanced.AllowRestartGame && DateTime.UtcNow.Subtract(LastRestartedGame).TotalMinutes >= 5)
            {
                LastRestartedGame = DateTime.UtcNow;
                string sUseProfile = Trinity.FirstProfile;
                Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Anti-stuck measures exiting current game.");
                // Load the first profile seen last run
                ProfileManager.Load(!string.IsNullOrEmpty(sUseProfile)
                                        ? sUseProfile
                                        : Zeta.Bot.ProfileManager.CurrentProfile.Path);
                Thread.Sleep(1000);
                Trinity.ResetEverythingNewGame();
                ZetaDia.Service.Party.LeaveGame(true);
                // Wait for 10 second log out timer if not in town
                if (!ZetaDia.IsInTown)
                {
                    Thread.Sleep(15000);
                }
            }
            else
            {
                Logger.Log(TrinityLogLevel.Info, LogCategory.UserInformation, "Unstucking measures failed. Now stopping Trinity unstucker for 12 minutes to inactivity timers to kick in or DB to auto-fix.");
                CancelUnstuckerForSeconds = 720;
                LastCancelledUnstucker = DateTime.UtcNow;
                return vSafeMovementLocation;
            }
            return vSafeMovementLocation;
        }
        // Handle moveto requests from the current routine/profile
        // This replaces DemonBuddy's own built-in "Basic movement handler" with a custom one
        private static Vector3 vLastMoveTo = Vector3.Zero;
        private static bool bLastWaypointWasTown = false;
        private static HashSet<Vector3> hashDoneThisVector = new HashSet<Vector3>();
        private static Vector3 vShiftedPosition = Vector3.Zero;
        private static DateTime lastShiftedPosition = DateTime.MinValue;

        private static Vector3 lastMovementPosition = Vector3.Zero;
        private static DateTime lastRecordedPosition = DateTime.UtcNow;

        internal static double MovementSpeed { get { return GetMovementSpeed(); } }

        private static List<SpeedSensor> SpeedSensors = new List<SpeedSensor>();
        private static int MaxSpeedSensors = 5;

        public static double GetMovementSpeed()
        {
            // Just changed worlds, Clean up the stack
            if (SpeedSensors.Any(s => s.WorldID != Trinity.Player.WorldDynamicID))
            {
                SpeedSensors.Clear();
                return 1d;
            }

            // record speed once per second
            if (DateTime.UtcNow.Subtract(lastRecordedPosition).TotalMilliseconds >= 1000)
            {
                // Record our current location and time
                if (!SpeedSensors.Any())
                {
                    SpeedSensors.Add(new SpeedSensor()
                    {
                        Location = MyPosition,
                        TimeSinceLastMove = new TimeSpan(0),
                        Distance = 0f,
                        WorldID = Trinity.Player.WorldDynamicID
                    });
                }
                else
                {
                    SpeedSensor lastSensor = SpeedSensors.OrderByDescending(s => s.Timestamp).FirstOrDefault();
                    SpeedSensors.Add(new SpeedSensor()
                    {
                        Location = MyPosition,
                        TimeSinceLastMove = new TimeSpan(DateTime.UtcNow.Subtract(lastSensor.TimeSinceLastMove).Ticks),
                        Distance = Vector3.Distance(MyPosition, lastSensor.Location),
                        WorldID = Trinity.Player.WorldDynamicID
                    });
                }

                lastRecordedPosition = DateTime.UtcNow;
            }

            // If we just used a spell, we "moved"
            if (DateTime.UtcNow.Subtract(Trinity.LastGlobalCooldownUse).TotalMilliseconds <= 1000)
                return 1d;

            if (DateTime.UtcNow.Subtract(Trinity.LastHadUnitInSights).TotalMilliseconds <= 1000)
                return 1d;

            if (DateTime.UtcNow.Subtract(Trinity.LastHadEliteUnitInSights).TotalMilliseconds <= 1000)
                return 1d;

            if (DateTime.UtcNow.Subtract(Trinity.LastHadContainerInSights).TotalMilliseconds <= 1000)
                return 1d;

            // Minimum of 2 records to calculate speed
            if (!SpeedSensors.Any() || SpeedSensors.Count <= 1)
                return 0d;

            // If we haven't "moved" in over a second, then we're standing still
            if (DateTime.UtcNow.Subtract(TimeLastUsedPlayerMover).TotalMilliseconds > 1000)
                return 0d;

            // Check if we have enough recorded positions, remove one if so
            while (SpeedSensors.Count > MaxSpeedSensors - 1)
            {
                // first sensors
                SpeedSensors.Remove(SpeedSensors.OrderBy(s => s.Timestamp).FirstOrDefault());
            }

            double AverageRecordingTime = SpeedSensors.Average(s => s.TimeSinceLastMove.TotalHours); ;
            double averageMovementSpeed = SpeedSensors.Average(s => Vector3.Distance(s.Location, MyPosition) * 1000000);

            return averageMovementSpeed / AverageRecordingTime;
        }

        /// <summary>
        /// Returns true if there's a blocking UIElement that we should NOT be moving!
        /// </summary>
        /// <returns></returns>
        public bool UISafetyCheck()
        {
            if (ElementIsVisible(UIElements.ConfirmationDialog))
                return true;
            if (ElementIsVisible(UIElements.ConfirmationDialogCancelButton))
                return true;
            if (ElementIsVisible(UIElements.ConfirmationDialogOkButton))
                return true;
            if (ElementIsVisible(UIElements.ReviveAtLastCheckpointButton))
                return true;

            return false;
        }

        private bool ElementIsVisible(UIElement element)
        {
            if (element == null)
                return false;
            if (!UIElement.IsValidElement(element.Hash))
                return false;
            if (!element.IsValid)
                return false;
            if (!element.IsVisible)
                return false;

            return true;
        }

        public int stuckAtt = 0;
        public Vector3 LastGeneratedFor = Vector3.Zero;
        public Vector3 LastGenerated = Vector3.Zero;

        public void MoveTowards(Vector3 moveToTarget)
        {
            if (Trinity.Settings.Advanced.DisableAllMovement)
                return;

            if (!Trinity.Player.IsFullyValid)
                return;

            if (UISafetyCheck())
                return;

            float distFromTarget = MyPosition.Distance2D(moveToTarget);

            if (!CacheData.NavigationObstacles.Any(o => o.Position.Distance2D(moveToTarget) <= 5) &&
                MainGrid.Map.Any() && distFromTarget > 5 && !Trinity.Player.IsInTown && stuckAtt < 5)
            {
                if (LastGeneratedFor.Distance2D(moveToTarget) <= 2 && LastGeneratedFor.Distance2D(MyPosition) >= 4)
                {
                    moveToTarget = LastGenerated;
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.Movement, "PlaverMover.ChangeToNode.RePeek:  MoveTowards={0}", NavHelper.PrettyPrintVector3(moveToTarget));
                    stuckAtt++;
                }
                else
                {
                    stuckAtt = 0;
                    string direction = MathUtil.GetHeadingToPoint(Trinity.Player.Position, moveToTarget);
                    var nodes = (
                        from n in MainGrid.Map
                        where
                            n.Distance <= 20 &&
                            n.Distance >= 5 &&
                            NavHelper.CanRayCast(n.Position)
                        orderby
                            NavHelper.CanRayCast(n.Position, moveToTarget) descending,
                            n.NearbyExitsWithinDistance(-1000, 8f) descending,
                            n.Position.Distance2D(moveToTarget)
                        select n).ToList();

                    if (nodes != null && nodes.Any() && nodes.Count() >= 1)
                    {
                        LastGeneratedFor = moveToTarget;
                        moveToTarget = nodes.First().Position;
                        LastGenerated = moveToTarget;

                        Logger.Log(TrinityLogLevel.Verbose, LogCategory.Movement, "PlaverMover.ChangeToNode: MoveTowards={0}", NavHelper.PrettyPrintVector3(moveToTarget));
                    }
                }
            } 

            Logger.Log(TrinityLogLevel.Verbose, LogCategory.Movement, "PlayerMover.MoveTowards={0}", NavHelper.PrettyPrintVector3(moveToTarget));

            TimeLastUsedPlayerMover = DateTime.UtcNow;
            LastMoveToTarget = moveToTarget;

            moveToTarget = WarnAndLogLongPath(moveToTarget);
            
            #region SpecialMovement
            if (UsedSpecialMovement(moveToTarget))
                return;

            // See if we can use abilities like leap etc. for movement out of combat, but not in town
            if (Trinity.Settings.Combat.Misc.AllowOOCMovement && !Trinity.Player.IsInTown && CombatBase.UseOOCPower)
            {
                // Leap movement for a barb
                if (Trinity.Settings.Combat.Barbarian.UseLeapOOC && CacheData.Hotbar.ActivePowers.Contains(SNOPower.Barbarian_Leap) &&
                    PowerManager.CanCast(SNOPower.Barbarian_Leap) && !ShrinesInArea(moveToTarget))
                {
                    Vector3 vThisTarget = moveToTarget;
                    if (distFromTarget > 35f)
                        vThisTarget = MathEx.CalculatePointFrom(moveToTarget, MyPosition, 35f);

                    if (CombatBase.Cast(new TrinityPower(SNOPower.Barbarian_Leap, 0f, vThisTarget)))
                    {
                        Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Using Leap for OOC movement, distance={0}", distFromTarget);
                        return;
                    }
                }
                // Furious Charge movement for a barb
                if (Trinity.Settings.Combat.Barbarian.UseChargeOOC && CacheData.Hotbar.ActivePowers.Contains(SNOPower.Barbarian_FuriousCharge) &&
                    distFromTarget >= 20f &&
                    PowerManager.CanCast(SNOPower.Barbarian_FuriousCharge) && !ShrinesInArea(moveToTarget))
                {
                    Vector3 vThisTarget = moveToTarget;
                    if (distFromTarget > 35f)
                        vThisTarget = MathEx.CalculatePointFrom(moveToTarget, MyPosition, 35f);

                    if (CombatBase.Cast(new TrinityPower(SNOPower.Barbarian_FuriousCharge, 0f, vThisTarget)))
                    {
                        Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Using Furious charge for OOC movement, distance={0}", distFromTarget);
                        return;
                    }
                } 

                // Tempest rush for a monk
                if (CacheData.Hotbar.ActivePowers.Contains(SNOPower.Monk_TempestRush) &&
                    (Trinity.Settings.Combat.Monk.TROption == TempestRushOption.MovementOnly || Trinity.Settings.Combat.Monk.TROption == TempestRushOption.Always ||
                    (Trinity.Settings.Combat.Monk.TROption == TempestRushOption.TrashOnly && !TargetUtil.AnyElitesInRange(40f))))
                {
                    Vector3 vTargetAimPoint = moveToTarget;

                    bool canRayCastTarget = true;

                    vTargetAimPoint = TargetUtil.FindTempestRushTarget();

                    if (!CanChannelTempestRush &&
                        ((Trinity.Player.PrimaryResource >= Trinity.Settings.Combat.Monk.TR_MinSpirit &&
                        distFromTarget >= Trinity.Settings.Combat.Monk.TR_MinDist) ||
                         DateTime.UtcNow.Subtract(CacheData.AbilityLastUsed[SNOPower.Monk_TempestRush]).TotalMilliseconds <= 150) &&
                        canRayCastTarget && PowerManager.CanCast(SNOPower.Monk_TempestRush))
                    {
                        CanChannelTempestRush = true;
                    }
                    else if ((CanChannelTempestRush && (Trinity.Player.PrimaryResource < 10f)) || !canRayCastTarget)
                    {
                        CanChannelTempestRush = false;
                    }

                    double lastUse = DateTime.UtcNow.Subtract(CacheData.AbilityLastUsed[SNOPower.Monk_TempestRush]).TotalMilliseconds;

                    if (CanChannelTempestRush)
                    {
                        if (Trinity.SNOPowerUseTimer(SNOPower.Monk_TempestRush))
                        {
                            LastTempestRushPosition = vTargetAimPoint;

                            if (CombatBase.Cast(new TrinityPower(SNOPower.Monk_TempestRush, 0f, vTargetAimPoint)))
                            {
                                // simulate movement speed of 30
                                SpeedSensor lastSensor = SpeedSensors.OrderByDescending(s => s.Timestamp).FirstOrDefault();
                                SpeedSensors.Add(new SpeedSensor()
                                {
                                    Location = MyPosition,
                                    TimeSinceLastMove = new TimeSpan(0, 0, 0, 0, 1000),
                                    Distance = 5f,
                                    WorldID = Trinity.Player.WorldDynamicID
                                });

                                if (Trinity.Settings.Advanced.LogCategories.HasFlag(LogCategory.Movement))
                                    Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Using Tempest Rush for OOC movement, distance={0:0} spirit={1:0} cd={2} lastUse={3:0} V3={4} vAim={5}",
                                        distFromTarget, Trinity.Player.PrimaryResource, PowerManager.CanCast(SNOPower.Monk_TempestRush), lastUse, moveToTarget, vTargetAimPoint);
                                return;
                            }
                        }
                    }
                    else
                    {
                        if (Trinity.Settings.Advanced.LogCategories.HasFlag(LogCategory.Movement))
                            Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement,
                                "Tempest rush failed!: {0:00.0} / {1} distance: {2:00.0} / {3} Raycast: {4} MS: {5:0.0} lastUse={6:0}",
                                Trinity.Player.PrimaryResource,
                                Trinity.Settings.Combat.Monk.TR_MinSpirit,
                                distFromTarget,
                                Trinity.Settings.Combat.Monk.TR_MinDist,
                                canRayCastTarget,
                                GetMovementSpeed(),
                                lastUse);
                    }
                }


                bool hasCalamity = CacheData.Hotbar.ActiveSkills.Any(s => s.Power == SNOPower.Wizard_Teleport && s.RuneIndex == 0);

                // Teleport for a wizard 
                if (!hasCalamity && CombatBase.CanCast(SNOPower.Wizard_Teleport, CanCastFlags.NoTimer) &&
                    CombatBase.TimeSincePowerUse(SNOPower.Wizard_Teleport) > 250 &&
                    distFromTarget >= 10f && !ShrinesInArea(moveToTarget))
                {
                    const float maxTeleportRange = 75f;

                    Vector3 vThisTarget = moveToTarget;
                    if (distFromTarget > maxTeleportRange)
                        vThisTarget = MathEx.CalculatePointFrom(moveToTarget, MyPosition, maxTeleportRange);

                    if (CombatBase.Cast(new TrinityPower(SNOPower.Wizard_Teleport, 0f, vThisTarget)))
                    {
                        Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Using Teleport for OOC movement, distance={0}", distFromTarget);
                        return;
                    }
                }
                // Archon Teleport for a wizard 
                if (CacheData.Hotbar.ActivePowers.Contains(SNOPower.Wizard_Archon_Teleport) &&
                    DateTime.UtcNow.Subtract(CacheData.AbilityLastUsed[SNOPower.Wizard_Archon_Teleport]).TotalMilliseconds >= CombatBase.GetSNOPowerUseDelay(SNOPower.Wizard_Archon_Teleport) &&
                    distFromTarget >= 20f &&
                    PowerManager.CanCast(SNOPower.Wizard_Archon_Teleport) && !ShrinesInArea(moveToTarget))
                {
                    Vector3 vThisTarget = moveToTarget;
                    if (distFromTarget > 35f)
                        vThisTarget = MathEx.CalculatePointFrom(moveToTarget, MyPosition, 35f);

                    if (CombatBase.Cast(new TrinityPower(SNOPower.Wizard_Archon_Teleport, 0f, vThisTarget)))
                    {
                        Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Using Archon Teleport for OOC movement, distance={0}", distFromTarget);
                        return;
                    }
                }
            }
            #endregion

            if (distFromTarget > 1f)
            {
                // Default movement
                ZetaDia.Me.UsePower(SNOPower.Walk, moveToTarget, Trinity.Player.WorldDynamicID, -1);

                Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Moved to:{0} dir:{1} Speed:{2:0.00} Dist:{3:0} ZDiff:{4:0} CanStand:{5} Raycast:{6}",
                    NavHelper.PrettyPrintVector3(moveToTarget), 
                    MathUtil.GetHeadingToPoint(moveToTarget), 
                    MovementSpeed, 
                    MyPosition.
                    Distance2D(moveToTarget),
                    MathUtil.GetDiff(MyPosition.Z, moveToTarget.Z),
                    Trinity.MainGridProvider.CanStandAt(Trinity.MainGridProvider.WorldToGrid(moveToTarget.ToVector2())),
                    NavHelper.CanRayCast(moveToTarget)
                    );
            }
            else
            {
                Logger.Log(TrinityLogLevel.Debug, LogCategory.Movement, "Reached MoveTowards Destination {0} Current Speed: {1:0.0}", moveToTarget, MovementSpeed);
            }
        }


        private static Vector3 WarnAndLogLongPath(Vector3 vMoveToTarget)
        {
            // The below code is to help profile/routine makers avoid waypoints with a long distance between them.
            // Long-distances between waypoints is bad - it increases stucks, and forces the DB nav-server to be called.
            if (Trinity.Settings.Advanced.LogStuckLocation)
            {
                if (vLastMoveTo == Vector3.Zero)
                    vLastMoveTo = vMoveToTarget;
                if (vMoveToTarget != vLastMoveTo)
                {
                    float fDistance = Vector3.Distance(vMoveToTarget, vLastMoveTo);
                    // Log if not in town, last waypoint wasn't FROM town, and the distance is >200 but <2000 (cos 2000+ probably means we changed map zones!)
                    if (!Trinity.Player.IsInTown && !bLastWaypointWasTown && fDistance >= 200 & fDistance <= 2000)
                    {
                        if (!hashDoneThisVector.Contains(vMoveToTarget))
                        {
                            // Log it
                            FileStream LogStream = File.Open(Path.Combine(FileManager.LoggingPath, "LongPaths - " + ZetaDia.Me.ActorClass.ToString() + ".log"), FileMode.Append, FileAccess.Write, FileShare.Read);
                            using (StreamWriter LogWriter = new StreamWriter(LogStream))
                            {
                                LogWriter.WriteLine(DateTime.UtcNow.ToString() + ":");
                                LogWriter.WriteLine("Profile Name=" + ProfileManager.CurrentProfile.Name);
                                LogWriter.WriteLine("'From' Waypoint=" + vLastMoveTo.ToString() + ". 'To' Waypoint=" + vMoveToTarget.ToString() + ". Distance=" + fDistance.ToString());
                            }
                            LogStream.Close();
                            hashDoneThisVector.Add(vMoveToTarget);
                        }
                    }
                    vLastMoveTo = vMoveToTarget;
                    bLastWaypointWasTown = false;
                    if (Trinity.Player.IsInTown)
                        bLastWaypointWasTown = true;
                }
            }
            return vMoveToTarget;
        }


        private static TrinityCacheObject CurrentTarget { get { return Trinity.CurrentTarget; } }

        private static Vector3 LastNavigateToTarget = Vector3.Zero;
        private static DateTime LastNavigateToResult = DateTime.MinValue;
        internal static MoveResult NavigateTo(Vector3 destination, string destinationName = "")
        {
            using (new MemorySpy("PlayerMover.NavigateTo()"))
            {
                PositionCache.AddPosition();
                MoveResult result;
                LastMoveToTarget = destination;

                if (NavHelper.CanRayCast(destination))
                {
                    Navigator.PlayerMover.MoveTowards(destination);
                    return MoveResult.Moved;
                }

                if (LastNavigateToTarget != destination && LastNavigateToTarget.Distance2D(destination) >= 3f)
                {
                    Navigator.Clear();
                    LastNavigateToResult = DateTime.MinValue;
                    LastNavigateToTarget = destination;
                }

                if (DateTime.UtcNow.Subtract(LastNavigateToResult).TotalMilliseconds < 350)
                    return MoveResult.Moved;

                try
                {
                    Stopwatch t1 = new Stopwatch();
                    t1.Start();

                    result = Navigator.MoveTo(LastNavigateToTarget, destinationName, false);

                    t1.Stop();

                    const float maxTime = 750;

                    // Shit was slow, make it slower but tell us why :)
                    string pathCheck = "";
                    if (Trinity.Settings.Advanced.LogCategories.HasFlag(LogCategory.Navigator) && t1.ElapsedMilliseconds > maxTime)
                    {
                        if (Navigator.GetNavigationProviderAs<DefaultNavigationProvider>().CanFullyClientPathTo(LastNavigateToTarget))
                            pathCheck = "CanFullyPath";
                        else
                            pathCheck = "CannotFullyPath";
                    }

                    LogCategory lc;
                    if (t1.ElapsedMilliseconds > maxTime)
                        lc = LogCategory.UserInformation;
                    else
                        lc = LogCategory.Navigator;

                    Logger.Log(TrinityLogLevel.Verbose, lc, "{0} in {1:0}ms {2} dist={3:0} {4}",
                        result, t1.ElapsedMilliseconds, destinationName, LastNavigateToTarget.Distance2D(Trinity.Player.Position), pathCheck);
                }
                catch (OutOfMemoryException)
                {
                    Logger.LogDebug("Navigator ran out of memory!");
                    return MoveResult.Failed;
                }
                catch (Exception ex)
                {
                    Logger.Log("{0}", ex);
                    return MoveResult.Failed;
                }

                LastNavigateToResult = DateTime.UtcNow;
                return result;
            }
        }

        internal static void SetKiteMovement(string post = "")
        {
            CombatBase.QueuedMovement.Queue(new QueuedMovement
            {
                Id = CurrentTarget.RActorGuid,
                Name = CurrentTarget != null ? CurrentTarget.InternalName : post,
                Infos = CurrentTarget != null ? CurrentTarget.Infos : "Avoidance/Kite movement",
                Destination = CurrentTarget != null ? CurrentTarget.Position : GridMap.GetBestMoveNode().Position,
                StopCondition = m =>
                    m.Destination == Vector3.Zero ||
                    m.Destination.Distance2D(Trinity.Player.Position) <= 1f ||
                    CurrentTarget == null
                ,
                Options = new QueuedMovementOptions
                {
                    FailureBlacklistSeconds = 1.5,
                    TimeBeforeBlocked = 1500,
                    Logging = LogLevel.Info,
                    AcceptableDistance = 3f,
                    Type = CurrentTarget != null && CurrentTarget.IsKite ? MoveType.Kite : MoveType.Avoidance,
                }
            });
        }

        internal static bool UsedSpecialMovement(Vector3 loc)
        {
            var node = GridMap.GetNodeAt(loc);
            if (node != null && node.Position.Distance2D(loc) <= 4f)
                return UsedSpecialMovement(node);

            return UsedSpecialMovement(new GridNode(loc));
        }

        internal static bool UsedSpecialMovement(GridNode node)
        {
            using (new MemorySpy("PlayerMover.UsedSpecialMovement()"))
            {
                if (node == null || node.Position == Vector3.Zero)
                    return false;

                #region Vault
                // COMBAT
                if (Trinity.Player.ActorClass == ActorClass.DemonHunter)
                {
                    if (DemonHunterCombat.CanCastCombatVaultMovement)
                    {
                        bool cMove = DemonHunterCombat.IsVaultAptCombatMovement(node.Position);
                        bool kMove = DemonHunterCombat.IsVaultAptKiteMovement(node.Position);
                        bool aMove = DemonHunterCombat.IsVaultAptAvoidanceMovement(node.Position);

                        if (cMove || kMove || aMove)
                        {
                            GridNode newNode = node;
                            if ((kMove || (aMove && !Trinity.Player.TryToAvoidProjectile)) && 
                                (node.Distance <= 33f || node.Distance >= 40f))
                            {
                                newNode = GridMap.GetBestMoveNode(35f, prioritizeDist: true);
                            }

                            if (newNode != null && newNode.Position != Vector3.Zero)
                            {
                                GridNode vaultNode = newNode;
                                vaultNode = GridMap.GetNodeAt(MathEx.CalculatePointFrom(newNode.Position, Trinity.Player.Position, 35f));

                                if (vaultNode != null && vaultNode.Position != Vector3.Zero)
                                {
                                    cMove = DemonHunterCombat.IsVaultAptCombatMovement(vaultNode.Position);
                                    kMove = DemonHunterCombat.IsVaultAptKiteMovement(vaultNode.Position);

                                    bool adequateMovement = cMove || ((aMove || kMove) && GridMap.GetWeightAtPlayer <= vaultNode.DynamicWeight);

                                    if (NavHelper.CanRayCast(vaultNode.Position) && adequateMovement)
                                    {
                                        if (CombatBase.Cast(new TrinityPower(SNOPower.DemonHunter_Vault, 0f, vaultNode.Position)))
                                        {
                                            DemonHunterCombat.LogVault(vaultNode.Position, cMove, kMove, aMove, vaultNode.WeightInfos);
                                            return true;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    // OOC
                    if (Trinity.Settings.Combat.Misc.AllowOOCMovement && CombatBase.UseOOCPower &&
                        CacheData.Hotbar.ActivePowers.Contains(SNOPower.DemonHunter_Vault) &&
                        DemonHunterCombat.CanCastOocVaultMovement && !ShrinesInArea(node.Position) &&
                        node.Position.Distance2D(Trinity.Player.Position) >= 35f && NavHelper.CanRayCast(node.Position) &&
                        !TargetUtil.LocOrPathInAoE(node.Position))
                    {

                        Vector3 _vVaultTarget = node.Position.Distance2D(Trinity.Player.Position) > 35f ?
                            MathEx.CalculatePointFrom(node.Position, Trinity.Player.Position, 35f) : node.Position;

                        if (CombatBase.Cast(new TrinityPower(SNOPower.DemonHunter_Vault, 0f, _vVaultTarget)))
                        {
                            DemonHunterCombat.LogVault(node.Position);
                            return true;
                        }
                            
                    }
                }
                #endregion

                #region DashingStrike
                if (Trinity.Player.ActorClass == ActorClass.Monk)
                {
                    if (CurrentTarget != null && MonkCombat.CanCastDashingStrike && !CurrentTarget.IsUnit && !CurrentTarget.IsAvoidance && node.Distance >= 15f &&
                        CombatBase.TimeSincePowerUse(SNOPower.X1_Monk_DashingStrike) > 1250 && CurrentTarget.IsInLineOfSight &&
                        CombatBase.Cast(new TrinityPower(SNOPower.X1_Monk_DashingStrike, 0f, CurrentTarget.Position)))
                    {
                        Logger.Log(TrinityLogLevel.Info, LogCategory.Movement, "Using DashingStrike for Target attempt, TargetType={0} Dist={1}", CurrentTarget.Type, node.Distance);
                        return true;
                    }

                    if (CombatBase.UseOOCPower && MonkCombat.CanCastDashingStrike && Trinity.Settings.Combat.Monk.UseDashingStrikeOOC && node.Distance >= 15f &&
                        CombatBase.TimeSincePowerUse(SNOPower.X1_Monk_DashingStrike) > 1000 && NavHelper.CanRayCast(node.Position) &&
                        CombatBase.Cast(new TrinityPower(SNOPower.X1_Monk_DashingStrike, 0f, node.Position)))
                    {
                        Logger.Log(TrinityLogLevel.Verbose, LogCategory.Movement, "Using DashingStrike for OOC movement, Dist={0}", node.Distance);
                        return true;
                    }
                }
                #endregion

                #region Whirlwind
                if ((!Trinity.Player.IsInTown || node.Distance > 15) &&
                    (CombatBase.UseCombatPower || CombatBase.UseOOCPower || CombatBase.UseAvoidancePower || node.Distance > 10) &&
                    Skills.Barbarian.Whirlwind.IsActive && Trinity.Player.PrimaryResource >= 10 &&
                    (TargetUtil.AnyMobsInRange(6f, false) || Trinity.Settings.Combat.Barbarian.KeepWhirlwind || Sets.BulKathossOath.IsEquipped || CombatBase.IsTaegukBuffWillExpire) &&
                    CombatBase.Cast(new TrinityPower(SNOPower.Barbarian_Whirlwind, 0f, node.Position)))
                {
                    Logger.Log(TrinityLogLevel.Verbose, LogCategory.Movement, "Using Whirlwind for movement, Dist={0}", node.Distance);
                    return true;
                }
                #endregion

                return false; 
            }
        }

        internal static bool IsMovementToTarget(Vector3 loc)
        {
            return CurrentTarget != null && loc.Distance(CurrentTarget.Position) < CurrentTarget.Distance;
        }

        internal static bool IsMovementToUnit(Vector3 loc)
        {
            return Trinity.ObjectCache != null && Trinity.ObjectCache.Any(o => o.IsUnit && o.Weight > 0 && loc.Distance(o.Position) < o.Distance);
        }
    }
}
