﻿using System.Threading.Tasks;
using Trinity.Helpers;
using Zeta.Bot;
using Zeta.Bot.Profile;
using Zeta.TreeSharp;
using Zeta.XmlEngine;
using Logger = Trinity.Technicals.Logger;

namespace Trinity.ProfileTags
{
    [XmlElement("InfernoStatus")]
    internal class InfernoStatusTag : ProfileBehavior
    {
        public InfernoStatusTag() { }

        private bool _isDone;
        public override bool IsDone
        {
            get { return !IsActiveQuestStep || _isDone; }
        }

        [XmlAttribute("runStatus")]
        public string RunStatus { get; set; }

        protected override Composite CreateBehavior()
        {
            return new ActionRunCoroutine(ret => MainCoroutine());
        }
         
        public async Task<bool> MainCoroutine()
        {
            if (Trinity.Player.IsDead)
                return true;

            if (RunStatus == "")
            {
                Logger.LogDebug("[Inferno] Error in RunStatus");
                _isDone = true;
                return true;
            }

            if (Inferno.RunStatus.Contains(RunStatus))
            {
                Logger.LogDebug("[Inferno] RunStatus is already \"" + RunStatus + "\"");
                _isDone = true;
                return true;
            }

            Logger.LogDebug("[Inferno] RunStatus \"" + Inferno.RunStatus + "\" switch to \"" + RunStatus + "\"");
            Inferno.RunStatus = RunStatus;

            switch (RunStatus)
            {
                case "StartRun":
                    Logger.LogDebug("[Inferno] Next Act=" + Inferno.CurrentAct);
                    Inferno.DebugRunStats();
                    InfernoUI.RefreshStatus();
                    Inferno.IncrementCount("Profile");
                    break;
                case "EndRun":
                    Logger.LogDebug("[Inferno] Add Act" + Inferno.CurrentAct + " to UsedAct");
                    Inferno.UsedAct.Add(Inferno.CurrentAct);
                    break;
                case "LeaveGame":
                    Inferno.DebugRunStats();
                    Logger.LogDebug("[Inferno] Reset all UsedAct");
                    Inferno.UsedAct.Clear();
                    break;
                case "InTown":
                    InfernoUI.RefreshCount("Key");
                    InfernoUI.RefreshStatus();
                    break;
            }

            _isDone = true;
            return _isDone;
        }

        public override void OnStart()
        {
            _isDone = false;
        }

        public override void ResetCachedDone()
        {
            _isDone = false;
        }
    }
}