﻿using System;
using System.ComponentModel;
using System.IO;
using System.Windows.Controls;
using System.Windows.Input;
using Trinity.Technicals;
using Zeta.Common.Xml;

namespace Trinity
{
    [Zeta.XmlEngine.XmlElement("ConfigSetting")]
    public class ConfigSetting : XmlSettings
    {
        private static ConfigSetting _Instance;
        public static ConfigSetting Instance
        {
            get { return _Instance ?? (_Instance = new ConfigSetting()); }
        }

        public ConfigSetting()
            : base(Path.Combine(FileManager.ConfigPath, "Config.xml"))
        {
        }

        #region Fields

        private string _SettingsFileName;
        private string _RenameTextBox;

        #endregion Fields

        #region Properties

        [Zeta.XmlEngine.XmlElement("SettingsFileName")]
        [DefaultValue("Default")]
        public string SettingsFileName
        {
            get
            {
                return _SettingsFileName;
            }
            set
            {
                if (_SettingsFileName != value)
                {
                    _SettingsFileName = value;
                    OnPropertyChanged("SettingsFileName");
                }

                if (_RenameTextBox != value)
                {
                    _RenameTextBox = value;
                    OnPropertyChanged("RenameTextBox");
                }
            }
        }

        [Zeta.XmlEngine.XmlElement("RenameTextBox")]
        [DefaultValue("Default")]
        public string RenameTextBox
        {
            get
            {
                return _RenameTextBox;
            }
            set
            {
                if (_RenameTextBox != value)
                {
                    _RenameTextBox = value;
                    OnPropertyChanged("RenameTextBox");
                }
            }
        }

        #endregion Properties
    }

    class ConfigForm
    {
        public static TextBox RenameTextBox = new TextBox();
        public static ListBox SettingFileList = new ListBox();

        public static void Manage()
        {
            RenameTextBox.Text = ConfigSetting.Instance.SettingsFileName;
            DirectoryInfo d = new DirectoryInfo(FileManager.TrinitySettingsPath);

            if (SettingFileList != null && SettingFileList.Items.Count >= 1)
                SettingFileList.Items.Clear();

            foreach (var file in d.GetFiles("*.xml"))
            {
                SettingFileList.Items.Add(file.Name.Replace(".xml", ""));
            }

            SettingFileList.SelectedValue = ConfigSetting.Instance.SettingsFileName;
        }
    }
}
